import base64
import os
import sys
import hashlib

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import ciphers, serialization
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.ciphers import algorithms, modes
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.hashes import SHA1
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from quarry.net.auth import PlayerPublicKey

backend = default_backend()

PY3 = sys.version_info > (3,)
_yggdrasil_key = None

class Cipher(object):
    def __init__(self):
        self.disable()

    def enable(self, key):
        cipher = ciphers.Cipher(
            algorithms.AES(key), modes.CFB8(key), backend=backend)
        self.encryptor = cipher.encryptor()
        self.decryptor = cipher.decryptor()

    def disable(self):
        self.encryptor = None
        self.decryptor = None

    def encrypt(self, data):
        if self.encryptor:
            return self.encryptor.update(data)
        else:
            return data

    def decrypt(self, data):
        if self.decryptor:
            return self.decryptor.update(data)
        else:
            return data


def make_keypair():
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=1024,
        backend=default_backend())


def make_server_id():
    data = os.urandom(10)
    if PY3:
        parts = ["%02x" % c for c in data]
    else:
        parts = ["%02x" % ord(c) for c in data]

    return "".join(parts)


def make_verify_token():
    return os.urandom(4)


def make_shared_secret():
    return os.urandom(16)


def make_digest(*data):
    sha1 = hashlib.sha1()
    for d in data:
        sha1.update(d)

    digest = int(sha1.hexdigest(), 16)
    if digest >> 39*4 & 0x8:
        return"-%x" % ((-digest) & (2**(40*4)-1))
    else:
        return "%x" % digest


def export_public_key(keypair):
    return keypair.public_key().public_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PublicFormat.SubjectPublicKeyInfo)


def import_public_key(data):
    return serialization.load_der_public_key(
        data=data,
        backend=default_backend())


def encrypt_secret(public_key, shared_secret):
    return public_key.encrypt(
        plaintext=shared_secret,
        padding=padding.PKCS1v15())


def decrypt_secret(keypair, data):
    return keypair.decrypt(
        ciphertext=data,
        padding=padding.PKCS1v15())


def get_yggdrasil_session_key():
    global _yggdrasil_key

    if _yggdrasil_key is not None:
        return _yggdrasil_key

    yggdrasil_key_path = os.path.abspath(os.path.join(
            os.path.dirname(__file__),
            "..",
            "data",
            "keys",
            "yggdrasil_session_pubkey.der"))
    yggdrasil_key_file = open(yggdrasil_key_path, "rb")
    _yggdrasil_key = import_public_key(yggdrasil_key_file.read())

    return _yggdrasil_key


# Verify 1.19 signature
def verify_mojang_v1_signature(data: PlayerPublicKey):
    # Need key in PEM format
    key_text = base64.encodebytes(data.key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)).decode('ISO-8859-1')
    e = "-----BEGIN RSA PUBLIC KEY-----\n" + key_text + "-----END RSA PUBLIC KEY-----\n"

    try:
        # Signature is timestamp as string + public key in PEM format
        get_yggdrasil_session_key().verify(data.signature, bytes(str(data.expiry) + e, 'ascii'), PKCS1v15(), SHA1())
        return True
    except InvalidSignature:
        return False


# Verify 1.19.1+ signature
def verify_mojang_v2_signature(data: PlayerPublicKey, uuid):
    if uuid is None:
        return False

    try:
        # Signature is uuid bytes + timestamp bytes + public key bytes
        key_bytes = data.key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)
        get_yggdrasil_session_key()\
            .verify(data.signature, uuid.bytes + data.expiry.to_bytes(8, 'big') + key_bytes, PKCS1v15(), SHA1())
        return True
    except InvalidSignature:
        return False



import collections
import functools
import gzip
import time
import zlib

from quarry.types.buffer import Buffer
from quarry.types.chunk import PackedArray

_kinds = {}
_ids = {}


# Base types ------------------------------------------------------------------

@functools.total_ordering
class _Tag(object):
    __slots__ = ('value',)

    def __init__(self, value):
        self.value = value

    @classmethod
    def from_bytes(cls, bytes):
        return cls.from_buff(Buffer(bytes))

    @classmethod
    def from_buff(cls, buff):
        raise NotImplementedError

    def to_bytes(self):
        raise NotImplementedError

    def to_obj(self):
        return self.value

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.value)

    def __eq__(self, other):
        return self.to_obj() == other.to_obj()

    def __lt__(self, other):
        return self.to_obj() < other.to_obj()


class _DataTag(_Tag):
    __slots__ = ()
    fmt = None

    @classmethod
    def from_buff(cls, buff):
        return cls(buff.unpack(cls.fmt))

    def to_bytes(self):
        return Buffer.pack(self.fmt, self.value)


class _ArrayTag(_Tag):
    __slots__ = ()
    width = None

    @classmethod
    def from_buff(cls, buff):
        length = buff.unpack('i')
        data = buff.read(length * (cls.width // 8))
        return cls(PackedArray.from_bytes(data, length, cls.width, cls.width))

    def to_bytes(self):
        data = self.value.to_bytes()
        data = Buffer.pack('i', len(data) // (self.width // 8)) + data
        return data

    def to_obj(self):
        return list(self.value)


# NBT tags --------------------------------------------------------------------

class TagByte(_DataTag):
    __slots__ = ()
    fmt = 'b'


class TagShort(_DataTag):
    __slots__ = ()
    fmt = 'h'


class TagInt(_DataTag):
    __slots__ = ()
    fmt = 'i'


class TagLong(_DataTag):
    __slots__ = ()
    fmt = 'q'


class TagFloat(_DataTag):
    __slots__ = ()
    fmt = 'f'


class TagDouble(_DataTag):
    __slots__ = ()
    fmt = 'd'


class TagString(_Tag):
    __slots__ = ()

    @classmethod
    def from_buff(cls, buff):
        string_length = buff.unpack('H')
        return cls(buff.read(string_length).decode('utf8'))

    def to_bytes(self):
        data = self.value.encode('utf8')
        return Buffer.pack('H', len(data)) + data


class TagByteArray(_ArrayTag):
    __slots__ = ()
    width = 8


class TagIntArray(_ArrayTag):
    __slots__ = ()
    width = 32


class TagLongArray(_ArrayTag):
    __slots__ = ()
    width = 64


class TagList(_Tag):
    __slots__ = ()

    @classmethod
    def from_buff(cls, buff):
        inner_kind_id, array_length = buff.unpack('bi')
        inner_kind = _kinds[inner_kind_id]
        return cls([inner_kind.from_buff(buff) for _ in range(array_length)])

    def to_bytes(self):
        if len(self.value) > 0:
            head = self.value[0]
        else:
            head = TagByte(0)

        return Buffer.pack('bi', _ids[type(head)], len(self.value)) + \
               b"".join(tag.to_bytes() for tag in self.value)

    def to_obj(self):
        return [tag.to_obj() for tag in self.value]


class TagCompound(_Tag):
    __slots__ = ()

    root = False
    preserve_order = False

    @classmethod
    def from_buff(cls, buff):
        if cls.preserve_order:
            value = collections.OrderedDict()
        else:
            value = {}

        while True:
            kind_id = buff.unpack('b')
            if kind_id == 0:
                return cls(value)
            kind = _kinds[kind_id]
            name = TagString.from_buff(buff).value
            tag = kind.from_buff(buff)
            value[name] = tag
            if cls.root:
                return cls(value)

    def to_bytes(self):
        string = b""
        for name, tag in self.value.items():
            string += Buffer.pack('b', _ids[type(tag)])
            string += TagString(name).to_bytes()
            string += tag.to_bytes()

        if len(self.value) == 0 or not self.root:
            string += Buffer.pack('b', 0)

        return string

    def to_obj(self):
        return dict((name, tag.to_obj()) for name, tag in self.value.items())

    def update(self, other_tag):
        for name, new_tag in other_tag.value.items():
            old_tag = self.value.get(name)

            if old_tag and not new_tag:
                del self.value[name]
            elif isinstance(old_tag, TagCompound) \
                    and isinstance(new_tag, TagCompound):
                self.value[name].update(new_tag)
            else:
                self.value[name] = new_tag


class TagRoot(TagCompound):
    __slots__ = ()
    root = True

    @classmethod
    def from_body(cls, body):
        return cls({"": body})

    @property
    def body(self):
        return self.value[""]


# Register tags ---------------------------------------------------------------

_kinds[0] = type(None)
_kinds[1] = TagByte
_kinds[2] = TagShort
_kinds[3] = TagInt
_kinds[4] = TagLong
_kinds[5] = TagFloat
_kinds[6] = TagDouble
_kinds[7] = TagByteArray
_kinds[8] = TagString
_kinds[9] = TagList
_kinds[10] = TagCompound
_kinds[11] = TagIntArray
_kinds[12] = TagLongArray
_ids.update({v: k for k, v in _kinds.items()})


# Files -----------------------------------------------------------------------

class NBTFile(object):
    root_tag = None

    def __init__(self, root_tag):
        self.root_tag = root_tag

    @classmethod
    def load(cls, path):
        with gzip.open(path, 'rb') as fd:
            return cls(TagRoot.from_bytes(fd.read()))

    def save(self, path):
        with gzip.open(path, 'wb') as fd:
            fd.write(self.root_tag.to_bytes())


class RegionFile(object):
    """
    Experimental support for the Minecraft world storage format (``.mca``).
    """
    def __init__(self, path):
        self.fd = open(path, "r+b")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fd.close()

    def close(self):
        """
        Closes the region file.
        """
        self.fd.close()

    def save_chunk(self, chunk):
        """
        Saves the given chunk, which should be a ``TagRoot``, to the region
        file.
        """

        # Compress chunk
        chunk_x = chunk.body.value["Level"].value["xPos"].value
        chunk_z = chunk.body.value["Level"].value["zPos"].value
        chunk = zlib.compress(chunk.to_bytes())
        chunk = Buffer.pack('IB', len(chunk), 2) + chunk
        chunk_length = 1 + (len(chunk) - 1) // 4096

        # Load extents
        extents = [(0, 2)]
        self.fd.seek(0)
        buff = Buffer(self.fd.read(4096))
        for idx in range(1024):
            z, x = divmod(idx, 32)
            entry = buff.unpack('I')
            offset, length = entry >> 8, entry & 0xFF
            if offset > 0 and not (x == chunk_x and z == chunk_z):
                extents.append((offset, length))
        extents.sort()
        extents.append((extents[-1][0] + extents[-1][1] + chunk_length, 0))

        # Compute new extent
        for idx in range(len(extents) - 1):
            start = extents[idx][0] + extents[idx][1]
            end = extents[idx+1][0]
            if (end - start) >= chunk_length:
                chunk_offset = start
                extents.insert(idx+1, (chunk_offset, chunk_length))
                break

        # Write extent header
        self.fd.seek(4 * (32 * chunk_z + chunk_x))
        self.fd.write(Buffer.pack(
            'I', (chunk_offset << 8) | (chunk_length & 0xFF)))

        # Write timestamp header
        self.fd.seek(4096 + 4 * (32 * chunk_z + chunk_x))
        self.fd.write(Buffer.pack('I', int(time.time())))

        # Write chunk
        self.fd.seek(4096 * chunk_offset)
        self.fd.write(chunk)

        # Truncate file
        self.fd.seek(4096 * extents[-1][0])
        self.fd.truncate()

    def load_chunk(self, chunk_x, chunk_z):
        """
        Loads the chunk at the given co-ordinates from the region file.
        The co-ordinates should range from 0 to 31. Returns a ``TagRoot``.
        """

        buff = Buffer()

        # Read extent header
        self.fd.seek(4 * (32 * chunk_z + chunk_x))
        buff.add(self.fd.read(4))
        entry = buff.unpack('I')
        chunk_offset, chunk_length = entry >> 8, entry & 0xFF
        if chunk_offset == 0:
            raise ValueError((chunk_x, chunk_z))

        # Read chunk
        self.fd.seek(4096 * chunk_offset)
        buff.add(self.fd.read(4096 * chunk_length))
        chunk = buff.read(buff.unpack('IB')[0])
        chunk = zlib.decompress(chunk)
        chunk = TagRoot.from_bytes(chunk)
        return chunk

    def load_chunk_section(self, chunk_x, chunk_y, chunk_z):
        """
        Loads the chunk section at the given co-ordinates from the region file.
        The co-ordinates should range from 0 to 31. Returns a ``TagRoot``.
        """

        chunk = self.load_chunk(chunk_x, chunk_z)
        sections = chunk.body.value["Level"].value["Sections"].value
        for section in sections:
            if section.value["Y"].value == chunk_y:
                return chunk, section

        raise ValueError((chunk_x, chunk_y, chunk_z))


# Debug -----------------------------------------------------------------------

def alt_repr(tag, level=0):
    """
    Returns a human-readable representation of a tag using the same format as
    used the NBT specification.
    """
    name = lambda kind: type(kind).__name__.replace("Tag", "TAG_")

    if isinstance(tag, _ArrayTag):
        return "%s%s: %d entries" % (
            "  " * level,
            name(tag),
            len(tag.value))

    elif isinstance(tag, TagList):
        return "%s%s: %d entries\n%s{\n%s\n%s}" % (
            "  " * level,
            name(tag),
            len(tag.value),
            "  " * level,
            "\n".join(alt_repr(tag, level+1) for tag in tag.value),
            "  " * level)

    elif isinstance(tag, TagRoot):
        return "\n".join(
                alt_repr(tag, level).replace(': ', '("%s"): ' % name, 1)
                for name, tag in tag.value.items())

    elif isinstance(tag, TagCompound):
        return "%s%s: %d entries\n%s{\n%s\n%s}" % (
            "  " * level,
            name(tag),
            len(tag.value),
            "  " * level,
            "\n".join(
                alt_repr(tag, level+1).replace(': ', '("%s"): ' % name, 1)
                for name, tag in tag.value.items()),
            "  " * level)

    elif isinstance(tag, TagString):
        return '%s%s: "%s"' % (
            "  " * level,
            name(tag),
            tag.value)

    else:
        return "%s%s: %r" % (
            "  " * level,
            name(tag),
            tag.value)



import functools
import json
import re
from typing import List

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.hashes import SHA256

from quarry.types.uuid import UUID


def _load_styles():
    data = {
        "0": "black",
        "1": "dark_blue",
        "2": "dark_green",
        "3": "dark_aqua",
        "4": "dark_red",
        "5": "dark_purple",
        "6": "gold",
        "7": "gray",
        "8": "dark_gray",
        "9": "blue",
        "a": "green",
        "b": "aqua",
        "c": "red",
        "d": "light_purple",
        "e": "yellow",
        "f": "white",
        "k": "obfuscated",
        "l": "bold",
        "m": "strikethrough",
        "n": "underline",
        "o": "italic",
        "r": "reset",
    }

    code_by_name = {}
    code_by_prop = {}
    for code, name in data.items():
        code_by_name[name] = code
        if code in "klmnor":
            if name == "underline":
                prop = "underlined"
            else:
                prop = name
            code_by_prop[prop] = code

    return code_by_name, code_by_prop


code_by_name, code_by_prop = _load_styles()


@functools.total_ordering
class Message(object):
    """
    Represents a Minecraft chat message.
    """
    def __init__(self, value):
        self.value = value

    @classmethod
    def from_buff(cls, buff):
        return cls(buff.unpack_json())

    def to_bytes(self):
        from quarry.types.buffer import Buffer

        return Buffer.pack_json(self.value)

    @classmethod
    def from_string(cls, string):
        return cls({'text': string})

    def to_string(self, strip_styles=True):
        """
        Minecraft uses a JSON format to represent chat messages; this method
        retrieves a plaintext representation, optionally including styles
        encoded using old-school chat codes (U+00A7 plus one character).
        """

        def parse(obj):
            if isinstance(obj, str):
                return obj
            if isinstance(obj, list):
                return "".join((parse(e) for e in obj))
            if isinstance(obj, dict):
                text = ""
                for prop, code in code_by_prop.items():
                    if obj.get(prop):
                        text += "\u00a7" + code
                if "color" in obj:
                    text += "\u00a7" + code_by_name[obj["color"]]
                if "translate" in obj:
                    text += obj["translate"]
                    if "with" in obj:
                        args = ", ".join((parse(e) for e in obj["with"]))
                        text += "{%s}" % args
                if "text" in obj:
                    text += obj["text"]
                if "extra" in obj:
                    text += parse(obj["extra"])
                return text

        text = parse(self.value)
        if strip_styles:
            text = self.strip_chat_styles(text)
        return text

    @classmethod
    def strip_chat_styles(cls, text):
        return re.sub("\u00A7.", "", text)

    def __eq__(self, other):
        return self.value == other.value

    def __lt__(self, other):
        return self.value < other.value

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return "<Message %r>" % str(self)


class LastSeenMessage(object):
    def __init__(self, sender: UUID, signature=None):
        self.sender = sender
        self.signature = signature

    def __eq__(self, other):
        if isinstance(other, LastSeenMessage):
            return self.sender == other.sender and self.signature == other.signature
        return NotImplemented


class SignedMessageHeader(object):
    """
    Represents the header of a signed minecraft chat message sent from a 1.19+ client
    Includes the sender UUID and optional signature of the preceding message
    """

    def __init__(self, sender: UUID, previous_signature: bytes = None):
        self.sender = sender
        self.previous_signature = previous_signature

    def __eq__(self, other):
        if isinstance(other, SignedMessageHeader):
            return self.sender == other.sender and self.previous_signature == other.previous_signature
        elif isinstance(other, LastSeenMessage):
            return self.sender == other.sender and self.previous_signature == other.signature
        return NotImplemented


class SignedMessageBody(object):
    """
    Represents the body of a signed minecraft chat message sent from a 1.19+ client
    Includes the message content, optional decorated message, timestamp and salt
    """

    def __init__(self, message: str, timestamp: int, salt: int, decorated_message: Message = None,
                 last_seen: List[LastSeenMessage] = None):
        self.message = message
        self.decorated_message = decorated_message
        self.timestamp = timestamp
        self.salt = salt

        if last_seen is None:
            last_seen = []

        self.last_seen = last_seen

    def digest(self):
        digest = hashes.Hash(hashes.SHA256())

        digest.update(self.salt.to_bytes(8, 'big'))  # Salt
        digest.update(int(self.timestamp / 1000).to_bytes(8, 'big'))  # Timestamp in seconds
        digest.update(self.message.encode("utf-8"))  # Message bytes
        digest.update((70).to_bytes(1, 'big'))  # Mojang adds a 70 byte after the message for some reason?

        if self.decorated_message is not None:
            digest.update(self.decorated_message.value.encode("utf-8"))  # FIXME: Test this

        for entry in self.last_seen:
            digest.update((70).to_bytes(1, 'big'))  # Mojang adds a 70 byte before each entry for some reason?
            digest.update(entry.sender.bytes)
            digest.update(entry.signature)

        return digest.finalize()

    def __eq__(self, other):
        if isinstance(other, SignedMessageBody):
            return self.message == other.message \
                   and self.decorated_message == other.decorated_message \
                   and self.timestamp == other.timestamp \
                   and self.salt == other.salt
        return NotImplemented


class SignedMessage(object):
    """
    Represents a signed minecraft chat message sent from a 1.19+ client
    Includes:
     - The message header, containing the sender UUID and optionally the previous message's signature,
     - The message body, containing the signed message, optional signed decorated message, timestamp and salt
     - The message signature, which may not be valid and can be checked with verify()
     - Optional unsigned message content
    """

    def __init__(self, header: SignedMessageHeader, signature: bytes, signature_version: int, body: SignedMessageBody,
                 unsigned_content: Message = None):
        if signature_version < 759:
            raise Exception("Signed messages are not supported below protocol version 759")

        self.header = header
        self.signature = signature
        self.signature_version = signature_version
        self.body = body
        self.unsigned_content = unsigned_content

    def verify(self, key: RSAPublicKey):
        data = b''

        if key is None or self.header.sender is None:
            return False

        # 1.19.1 +
        if self.signature_version >= 760:
            if self.header.previous_signature is not None:
                data = data + self.header.previous_signature

            data = data + self.header.sender.bytes + self.body.digest()

        # 1.19
        else:
            data = data + self.body.salt.to_bytes(8, 'big') \
                   + self.header.sender.bytes \
                   + int(self.body.timestamp / 1000).to_bytes(8, 'big') \
                   + json.dumps(Message.from_string(self.body.message).value, sort_keys=True, separators=(',', ':')).encode('utf-8')

        try:
            key.verify(self.signature, data, PKCS1v15(), SHA256())
            return True
        except InvalidSignature:
            return False

    def __eq__(self, other):
        if isinstance(other, SignedMessage):
            return self.header == other.header \
                   and self.body == other.body \
                   and self.signature == other.signature \
                   and self.signature_version == other.signature_version \
                   and self.unsigned_content == other.unsigned_content
        return NotImplemented



import glob
import os.path
import re

from quarry.types.nbt import NBTFile, TagRoot


def _load():
    data_packs = {}
    dimension_types = {}
    nbt_paths = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        "data_packs",
        "*.nbt"))
    for nbt_path in glob.glob(nbt_paths):
        match = re.match('(\d{4})_(.+)\.nbt', os.path.basename(nbt_path))
        if not match:
            continue

        protocol_version = int(match.group(1))
        minecraft_version = match.group(2)
        data_pack = NBTFile.load(nbt_path).root_tag

        data_packs[protocol_version] = data_pack

        for entry in data_pack.body.value.values():
            if entry.value['type'].value == 'minecraft:dimension_type':
                for dimension in entry.value['value'].value:
                    name = dimension.value['name'].value
                    value = TagRoot.from_body(dimension.value['element'])
                    dimension_types[protocol_version, name] = value

    return data_packs, dimension_types


data_packs, dimension_types = _load()



from quarry.types.buffer.v1_9 import Buffer1_9
from quarry.types.registry import OpaqueRegistry


class Buffer1_13(Buffer1_9):
    registry = OpaqueRegistry(14)

    # Chunk section -----------------------------------------------------------

    @classmethod
    def pack_chunk_section_palette(cls, palette):
        if not palette:
            return b""
        else:
            return cls.pack_varint(len(palette)) + b"".join(
                cls.pack_varint(x) for x in palette)

    def unpack_chunk_section_palette(self, value_width):
        if value_width > 8:
            return []
        else:
            return [self.unpack_varint() for _ in range(self.unpack_varint())]

    # Slot --------------------------------------------------------------------

    @classmethod
    def pack_slot(cls, item=None, count=1, tag=None):
        """
        Packs a slot.
        """

        if item is None:
            return cls.pack('h', -1)

        item_id = cls.registry.encode('minecraft:item', item)
        return cls.pack('hb', item_id, count) + cls.pack_nbt(tag)

    def unpack_slot(self):
        """
        Unpacks a slot.
        """

        slot = {}
        item_id = self.unpack('h')
        if item_id == -1:
            slot['item'] = None
        else:
            slot['item'] = self.registry.decode('minecraft:item', item_id)
            slot['count'] = self.unpack('b')
            slot['tag'] = self.unpack_nbt()
        return slot

    # Entity metadata ---------------------------------------------------------

    @classmethod
    def pack_entity_metadata(cls, metadata):
        """
        Packs entity metadata.
        """

        pack_position = lambda pos: cls.pack_position(*pos)
        out = b""
        for ty_key, val in metadata.items():
            ty, key = ty_key
            out += cls.pack('BB', key, ty)
            if   ty == 0:  out += cls.pack('b', val)
            elif ty == 1:  out += cls.pack_varint(val)
            elif ty == 2:  out += cls.pack('f', val)
            elif ty == 3:  out += cls.pack_string(val)
            elif ty == 4:  out += cls.pack_chat(val)
            elif ty == 5:  out += cls.pack_optional(cls.pack_chat, val)
            elif ty == 6:  out += cls.pack_slot(**val)
            elif ty == 7:  out += cls.pack('?', val)
            elif ty == 8:  out += cls.pack_rotation(*val)
            elif ty == 9:  out += cls.pack_position(*val)
            elif ty == 10: out += cls.pack_optional(pack_position, val)
            elif ty == 11: out += cls.pack_direction(val)
            elif ty == 12: out += cls.pack_optional(cls.pack_uuid, val)
            elif ty == 13: out += cls.pack_block(val)
            elif ty == 14: out += cls.pack_nbt(val)
            elif ty == 15: out += cls.pack_particle(*val)
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
        out += cls.pack('B', 255)
        return out

    def unpack_entity_metadata(self):
        """
        Unpacks entity metadata.
        """

        metadata = {}
        while True:
            key = self.unpack('B')
            if key == 255:
                return metadata
            ty = self.unpack('B')
            if   ty == 0:  val = self.unpack('b')
            elif ty == 1:  val = self.unpack_varint()
            elif ty == 2:  val = self.unpack('f')
            elif ty == 3:  val = self.unpack_string()
            elif ty == 4:  val = self.unpack_chat()
            elif ty == 5:  val = self.unpack_optional(self.unpack_chat)
            elif ty == 6:  val = self.unpack_slot()
            elif ty == 7:  val = self.unpack('?')
            elif ty == 8:  val = self.unpack_rotation()
            elif ty == 9:  val = self.unpack_position()
            elif ty == 10: val = self.unpack_optional(self.unpack_position)
            elif ty == 11: val = self.unpack_direction()
            elif ty == 12: val = self.unpack_optional(self.unpack_uuid)
            elif ty == 13: val = self.unpack_block()
            elif ty == 14: val = self.unpack_nbt()
            elif ty == 15: val = self.unpack_particle()
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
            metadata[ty, key] = val

    # Particle ----------------------------------------------------------------

    @classmethod
    def pack_particle(cls, id, data=None):
        """
        Packs a particle.
        """

        data = data or {}
        out = cls.pack_varint(id)
        if id == 3 or id == 20:
            out += cls.pack_varint(data['block_state'])
        elif id == 11:
            out += cls.pack(
                'ffff',
                data['red'],
                data['green'],
                data['blue'],
                data['scale'])
        elif id == 27:
            out += cls.pack_slot(**data['item'])

        return out

    def unpack_particle(self):
        """
        Unpacks a particle. Returns an ``(id, data)`` pair.
        """

        id = self.unpack_varint()
        if id == 3 or id == 20:
            data = {'block_state': self.unpack_varint()}
        elif id == 11:
            data = dict(zip(
                ('red', 'green', 'blue', 'scale'),
                self.unpack('ffff')))
        elif id == 27:
            data = {'item': self.unpack_slot()}
        else:
            data = {}

        return id, data

    # Commands ----------------------------------------------------------------

    def unpack_commands(self, resolve_redirects=True):
        """
        Unpacks a command graph.

        If *resolve_redirects* is ``True`` (the default), the returned
        structure may contain contain circular references, and therefore cannot
        be serialized to JSON (or similar). If it is ``False``, all node
        redirect information is stripped, resulting in a directed acyclic
        graph.
        """

        # Unpack nodes
        node_count = self.unpack_varint()
        nodes = [self.unpack_command_node() for _ in range(node_count)]

        # Resolve children and redirects
        for node in nodes:
            node['children'] = {nodes[idx]['name']: nodes[idx]
                                for idx in node['children']}
            if node['redirect'] is not None:
                if resolve_redirects:
                    node['redirect'] = nodes[node['redirect']]
                else:
                    node['redirect'] = None

        return nodes[self.unpack_varint()]

    def unpack_command_node(self):
        """
        Unpacks a command node.
        """

        node = {}

        flags = self.unpack('B')
        node['type'] = ['root', 'literal', 'argument'][flags & 0x03]
        node['executable'] = bool(flags & 0x04)
        node['children'] = [self.unpack_varint() for _ in
                            range(self.unpack_varint())]
        node['redirect'] = self.unpack_varint() if flags & 0x08 else None
        node['name'] = self.unpack_string() if node['type'] != 'root' else None

        if node['type'] == 'argument':
            node['parser'] = self.unpack_string()
            node['properties'] = self.unpack_command_node_properties(node['parser'])

        node['suggestions'] = self.unpack_string() if flags & 0x10 else None

        return node

    def unpack_command_node_properties(self, parser):
        """
        Unpacks the properties of an ``argument`` command node.
        """

        namespace, parser = parser.split(":", 1)
        properties = {}

        if namespace == "brigadier":
            if parser == "bool":
                pass
            elif parser == "string":
                properties['behavior'] = self.unpack_varint()
            elif parser in ("double", "float", "integer"):
                fmt = parser[0]
                flags = self.unpack('B')
                properties['min'] = self.unpack(fmt) if flags & 0x01 else None
                properties['max'] = self.unpack(fmt) if flags & 0x02 else None

        elif namespace == "minecraft":
            if parser in ('entity', 'score_holder'):
                properties['allow_multiple'] = self.unpack('?')

            elif parser == 'range':
                properties['allow_decimals'] = self.unpack('?')

        return properties

    @classmethod
    def pack_commands(cls, root_node):
        """
        Packs a command graph.
        """

        # Enumerate nodes
        nodes = [root_node]
        idx = 0
        while idx < len(nodes):
            node = nodes[idx]
            children = list(node['children'].values())
            if node['redirect']:
                children.append(node['redirect'])

            for child in children:
                if child not in nodes:
                    nodes.append(child)
            idx += 1

        # Pack nodes
        out = cls.pack_varint(len(nodes))
        for node in nodes:
            out += cls.pack_command_node(node, nodes)

        out += cls.pack_varint(nodes.index(root_node))

        return out

    @classmethod
    def pack_command_node(cls, node, nodes):
        """
        Packs a command node.
        """

        out = b""

        flags = (
            ['root', 'literal', 'argument'].index(node['type']) |
            int(node['executable']) << 2 |
            int(node['redirect'] is not None) << 3 |
            int(node['suggestions'] is not None) << 4)
        out += cls.pack('B', flags)
        out += cls.pack_varint(len(node['children']))

        for child in node['children'].values():
            out += cls.pack_varint(nodes.index(child))

        if node['redirect'] is not None:
            out += cls.pack_varint(nodes.index(node['redirect']))

        if node['name'] is not None:
            out += cls.pack_string(node['name'])

        if node['type'] == 'argument':
            out += cls.pack_string(node['parser'])
            out += cls.pack_command_node_properties(node['parser'],
                                                    node['properties'])
        if node['suggestions'] is not None:
            out += cls.pack_string(node['suggestions'])

        return out

    @classmethod
    def pack_command_node_properties(cls, parser, properties):
        """
        Packs the properties of an ``argument`` command node.
        """

        namespace, parser = parser.split(":", 1)
        out = b""

        if namespace == "brigadier":
            if parser == "bool":
                pass
            elif parser == "string":
                out += cls.pack_varint(properties['behavior'])
            elif parser in ("double", "float", "integer"):
                fmt = parser[0]
                flags = (
                    int(properties['min'] is not None) |
                    int(properties['max'] is not None) << 1)
                out += cls.pack('B', flags)
                if properties['min'] is not None:
                    out += cls.pack(fmt, properties['min'])
                if properties['max'] is not None:
                    out += cls.pack(fmt, properties['max'])

        elif namespace == "minecraft":
            if parser in ('entity', 'score_holder'):
                out += cls.pack('?', properties['allow_multiple'])

            elif parser == 'range':
                out += cls.pack('?', properties['allow_decimals'])

        return out

    # Recipes -----------------------------------------------------------------

    def unpack_recipe(self):
        """
        Unpacks a crafting recipe.
        """
        recipe = {}
        recipe['name'] = self.unpack_string()
        recipe['type'] = self.unpack_string()

        if recipe['type'] == 'crafting_shapeless':
            recipe['group'] = self.unpack_string()
            recipe['ingredients'] = [
                self.unpack_ingredient() for _ in range(self.unpack_varint())]
            recipe['result'] = self.unpack_slot()

        elif recipe['type'] == 'crafting_shaped':
            recipe['width'] = self.unpack_varint()
            recipe['height'] = self.unpack_varint()
            recipe['group'] = self.unpack_string()
            recipe['ingredients'] = [
                self.unpack_ingredient() for _ in range(recipe['width'] *
                                                    recipe['height'])]
            recipe['result'] = self.unpack_slot()
        elif recipe['type'] == 'smelting':
            recipe['group'] = self.unpack_string()
            recipe['ingredient'] = self.unpack_ingredient()
            recipe['result'] = self.unpack_slot()
            recipe['experience'] = self.unpack('f')
            recipe['cooking_time'] = self.unpack_varint()

        return recipe

    @classmethod
    def pack_recipe(cls, name, type, **recipe):
        """
        Packs a crafting recipe.
        """
        data = cls.pack_string(name) + cls.pack_string(type)

        if type == 'crafting_shapeless':
            data += cls.pack_string(recipe['group'])
            data += cls.pack_varint(len(recipe['ingredients']))
            for ingredient in recipe['ingredients']:
                data += cls.pack_ingredient(ingredient)
            data += cls.pack_slot(**recipe['result'])

        elif type == 'crafting_shaped':
            data += cls.pack_varint(recipe['width'])
            data += cls.pack_varint(recipe['height'])
            data += cls.pack_string(recipe['group'])
            for ingredient in recipe['ingredients']:
                data += cls.pack_ingredient(ingredient)
            data += cls.pack_slot(**recipe['result'])

        elif type == 'smelting':
            data += cls.pack_string(recipe['group'])
            data += cls.pack_ingredient(recipe['ingredient'])
            data += cls.pack_slot(**recipe['result'])
            data += cls.pack('f', recipe['experience'])
            data += cls.pack_varint(recipe['cooking_time'])

        return data

    def unpack_ingredient(self):
        """
        Unpacks a crafting recipe ingredient alternation.
        """
        return [self.unpack_slot() for _ in range(self.unpack_varint())]

    @classmethod
    def pack_ingredient(cls, ingredient):
        """
        Packs a crafting recipe ingredient alternation.
        """
        data = cls.pack_varint(len(ingredient))
        for slot in ingredient:
            data += cls.pack_slot(**slot)
        return data



import logging
from twisted.internet import protocol

from quarry.data import packets
from quarry.types.buffer import BufferUnderrun, buff_types
from quarry.net.crypto import Cipher
from quarry.net.ticker import Ticker

protocol_modes = {
    0: 'init',
    1: 'status',
    2: 'login',
    3: 'play'
}
protocol_modes_inv = dict(((v, k) for k, v in protocol_modes.items()))


class ProtocolError(Exception):
    pass


class PacketDispatcher(object):
    def dispatch(self, lookup_args, buff):
        handler = getattr(self, f"packet_{'_'.join(lookup_args)}", None)
        if handler is not None:
            handler(buff)
            return True
        return False


class Protocol(protocol.Protocol, PacketDispatcher, object):
    """Shared logic between the client and server"""

    #: Usually a reference to a :class:`~quarry.types.buffer.Buffer` class.
    #: This is useful when constructing a packet payload for use in
    #: :meth:`send_packet`
    buff_type = None

    #: The logger for this protocol.
    logger = None

    #: A reference to a :class:`~quarry.net.ticker.Ticker` instance.
    ticker = None

    #: A reference to the factory
    factory = None

    #: The IP address of the remote.
    remote_addr = None

    recv_direction = None
    send_direction = None
    protocol_version = packets.default_protocol_version
    protocol_mode = "init"
    compression_threshold = -1
    in_game = False
    closed = False

    def __init__(self, factory, remote_addr):
        self.factory = factory
        self.remote_addr = remote_addr

        self.buff_type = self.factory.get_buff_type(self.protocol_version)
        self.recv_buff = self.buff_type()
        self.cipher = Cipher()

        self.logger = logging.getLogger("%s{%s}" % (
            self.__class__.__name__,
            self.remote_addr.host))
        self.logger.setLevel(self.factory.log_level)

        self.ticker = self.factory.ticker_type(self.logger)
        self.ticker.start()

        self.connection_timer = self.ticker.add_delay(
            delay=self.factory.connection_timeout / self.ticker.interval,
            callback=self.connection_timed_out)

        self.setup()

    # Fix ugly twisted methods ------------------------------------------------

    def dataReceived(self, data):
        return self.data_received(data)

    def connectionMade(self):
        return self.connection_made()

    def connectionLost(self, reason=None):
        return self.connection_lost(reason)

    # Convenience functions ---------------------------------------------------

    def check_protocol_mode_switch(self, mode):
        transitions = [
            ("init", "status"),
            ("init", "login"),
            ("login", "play")
        ]

        if (self.protocol_mode, mode) not in transitions:
            raise ProtocolError("Cannot switch protocol mode from %s to %s"
                                % (self.protocol_mode, mode))

    def switch_protocol_mode(self, mode):
        self.check_protocol_mode_switch(mode)
        self.protocol_mode = mode

    def set_compression(self, compression_threshold):
        self.compression_threshold = compression_threshold
        self.logger.debug("Compression threshold set to %d bytes"
                          % compression_threshold)

    def close(self, reason=None):
        """Closes the connection"""

        if not self.closed:
            if reason:
                reason = "Closing connection: %s" % reason
            else:
                reason = "Closing connection"

            if self.in_game:
                self.logger.info(reason)
            else:
                self.logger.debug(reason)

            self.transport.loseConnection()
            self.closed = True

    def log_packet(self, prefix, name):
        """Logs a packet at debug level"""

        self.logger.debug("Packet %s %s/%s" % (
            prefix,
            self.protocol_mode,
            name))

    # General callbacks -------------------------------------------------------

    def setup(self):
        """Called when the Protocol's initialiser is finished"""

        pass

    def protocol_error(self, err):
        """Called when a protocol error occurs"""

        self.logger.exception(err)
        self.close("Protocol error")

    # Connection callbacks ----------------------------------------------------

    def connection_made(self):
        """Called when the connection is established"""

        self.logger.debug("Connection made")

    def connection_lost(self, reason=None):
        """Called when the connection is lost"""

        self.closed = True
        if self.in_game:
            self.player_left()
        self.logger.debug("Connection lost")

        self.ticker.stop()

    def connection_timed_out(self):
        """Called when the connection has been idle too long"""

        self.close("Connection timed out")

    # Auth callbacks ----------------------------------------------------------

    def auth_ok(self, data):
        """Called when auth with mojang succeeded (online mode only)"""

        pass

    def auth_failed(self, err):
        """Called when auth with mojang failed (online mode only)"""

        self.logger.warning("Auth failed: %s" % err.value)
        self.close("Auth failed: %s" % err.value)

    # Player callbacks --------------------------------------------------------

    def player_joined(self):
        """Called when the player joins the game"""

        self.in_game = True

    def player_left(self):
        """Called when the player leaves the game"""

        pass

    # Packet handling ---------------------------------------------------------

    def get_packet_name(self, ident):
        key = (self.protocol_version, self.protocol_mode, self.recv_direction,
               ident)
        try:
            return packets.packet_names[key]
        except KeyError:
            raise ProtocolError("No name known for packet: %s" % (key,))

    def get_packet_ident(self, name):
        key = (self.protocol_version, self.protocol_mode, self.send_direction,
               name)
        try:
            return packets.packet_idents[key]
        except KeyError:
            raise ProtocolError("No ID known for packet: %s" % (key,))

    def data_received(self, data):
        # Decrypt data
        data = self.cipher.decrypt(data)

        # Add it to our buffer
        self.recv_buff.add(data)

        # Read some packets
        while not self.closed:
            # Save the buffer, in case we read an incomplete packet
            self.recv_buff.save()

            # Read the packet
            try:
                buff = self.recv_buff.unpack_packet(
                    self.buff_type,
                    self.compression_threshold)

            except BufferUnderrun:
                self.recv_buff.restore()
                break

            try:
                # Identify the packet
                name = self.get_packet_name(buff.unpack_varint())

                # Dispatch the packet
                try:
                    self.packet_received(buff, name)
                except BufferUnderrun:
                    raise ProtocolError("Packet is too short: %s" % name)
                if len(buff) > 0:
                    raise ProtocolError("Packet is too long: %s" % name)

                # Reset the inactivity timer
                self.connection_timer.restart()

            except ProtocolError as e:
                self.protocol_error(e)

    def packet_received(self, buff, name):
        """
        Called when a packet is received from the remote. Usually this method
        dispatches the packet to a method named ``packet_<packet name>``, or
        calls :meth:`packet_unhandled` if no such methods exists. You might
        want to override this to implement your own dispatch logic or logging.
        """

        self.log_packet(". recv", name)

        dispatched = self.dispatch((name,), buff)

        if not dispatched:
            self.packet_unhandled(buff, name)

    def packet_unhandled(self, buff, name):
        """
        Called when a packet is received that is not hooked. The default
        implementation silently discards the packet.
        """

        buff.discard()

    def send_packet(self, name, *data):
        """Sends a packet to the remote."""

        if self.closed:
            return

        self.log_packet("# send", name)

        data = b"".join(data)

        # Prepend ident
        data = self.buff_type.pack_varint(self.get_packet_ident(name)) + data

        # Pack packet
        data = self.buff_type.pack_packet(data, self.compression_threshold)

        # Encrypt
        data = self.cipher.encrypt(data)

        # Send
        self.transport.write(data)


class Factory(protocol.Factory, object):
    protocol = Protocol
    ticker_type = Ticker
    log_level = logging.INFO
    connection_timeout = 30
    force_protocol_version = None

    minecraft_versions = packets.minecraft_versions

    def buildProtocol(self, addr):
        return self.protocol(self, addr)

    def get_buff_type(self, protocol_version):
        """
        Gets a buffer type for the given protocol version.
        """
        for ver, cls in reversed(buff_types):
            if protocol_version >= ver:
                return cls



from quarry.types.buffer.v1_13 import Buffer1_13


class Buffer1_13_2(Buffer1_13):
    @classmethod
    def pack_slot(cls, item=None, count=1, tag=None):
        """
        Packs a slot.
        """

        if item is None:
            return cls.pack('?', False)
        else:
            return cls.pack('?', True) + \
                   cls.pack_varint(
                       cls.registry.encode('minecraft:item', item)) + \
                   cls.pack('b', count) + \
                   cls.pack_nbt(tag)

    def unpack_slot(self):
        """
        Unpacks a slot.
        """

        slot = {}
        item_id = self.unpack_optional(self.unpack_varint)
        if item_id is None:
            slot['item'] = None
        else:
            slot['item'] = self.registry.decode('minecraft:item', item_id)
            slot['count'] = self.unpack('b')
            slot['tag'] = self.unpack_nbt()
        return slot


class BufferUnderrun(Exception):
    pass


from quarry.types.buffer.v1_7 import Buffer1_7
from quarry.types.buffer.v1_9 import Buffer1_9
from quarry.types.buffer.v1_13 import Buffer1_13
from quarry.types.buffer.v1_13_2 import Buffer1_13_2
from quarry.types.buffer.v1_14 import Buffer1_14
from quarry.types.buffer.v1_19 import Buffer1_19
from quarry.types.buffer.v1_19_1 import Buffer1_19_1


# Versioned buffers used after handshaking
buff_types = [
    (0,   Buffer1_7),
    (107, Buffer1_9),
    (393, Buffer1_13),
    (404, Buffer1_13_2),
    (477, Buffer1_14),
    (759, Buffer1_19),
    (760, Buffer1_19_1),
]


# Used by NBT and during handshaking
class Buffer(buff_types[-1][1]):
    pass



import json
import math
import os.path
import subprocess


class Registry(object):
    """
    Base type for registries.
    """

    #: Number of bits needed to represent the greatest block ID.
    max_bits = None


    def encode(self, kind, obj):
        """
        Encodes a thing to an integer ID.
        """

        raise NotImplementedError

    def decode(self, kind, val):
        """
        Decodes a thing from an integer ID.
        """

        raise NotImplementedError

    def encode_block(self, obj):
        """
        Encodes a block to an integer ID.
        """

        raise NotImplementedError

    def decode_block(self, val):
        """
        Decodes a block from an integer ID.
        """

        raise NotImplementedError

    def is_air_block(self, obj):
        """
        Returns true if the given object is considered air for lighting
        purposes.
        """

        raise NotImplementedError

class OpaqueRegistry(Registry):
    """
    Registry that passes IDs through unchanged. This is the default.
    """

    def __init__(self, max_bits):
        self.max_bits = max_bits


    def encode(self, kind, obj): return obj
    def decode(self, kind, val): return val

    def encode_block(self, obj): return obj
    def decode_block(self, val): return val
    def is_air_block(self, obj): return obj == 0


class BitShiftRegistry(OpaqueRegistry):
    """
    Registry implementing the Minecraft 1.7 - 1.12 bit-shift format for blocks.

    Blocks decode to a ``(block_id, metadata)`` pair. Items pass through
    unchanged.
    """

    max_bits = 13

    def encode_block(self, obj): return (obj[0] << 4) | obj[1]
    def decode_block(self, val): return val >> 4, val & 0x0F
    def is_air_block(self, obj): return obj[0] == 0


class LookupRegistry(Registry):
    """
    Registry implementing a dictionary lookup, recommended for 1.13+.

    Blocks decode to a ``dict`` where the only guaranteed key is ``'name'``.
    Items decode to a ``str`` name.

    Use the ``from_jar()`` or ``from_json()`` class methods to load data from
    the official server.
    """


    def __init__(self, blocks, registries):
        self.max_bits = int(math.ceil(math.log(max(blocks.keys()), 2)))

        self.decode_block_map = blocks
        self.encode_block_map = {
            frozenset(value.items()): key
            for key, value in blocks.items()}

        self.decode_map = registries
        self.encode_map = {
            registry_name: {value: key for key, value in registry.items()}
            for registry_name, registry in registries.items()}

    def encode(self, registry, obj):
        return self.encode_map[registry][obj]

    def decode(self, registry, val):
        return self.decode_map[registry][val]

    def encode_block(self, obj):
        return self.encode_block_map[frozenset(obj.items())]

    def decode_block(self, val):
        return dict(self.decode_block_map[val])

    def is_air_block(self, obj):
        for name in ('air', 'cave_air', 'void_air'):
            if obj['name'] == name:
                return True
        return False

    @classmethod
    def from_jar(cls, jar_path):
        """
        Create a ``LookupRegistry`` from a Minecraft server jar file. This
        method generates JSON files by running the Minecraft server like so::

            java -cp minecraft_server.jar net.minecraft.data.Main --reports

        It then feeds the generated JSON files to ``from_json()``.
        """

        root_path, jar_name = os.path.split(jar_path)
        reports_path = os.path.join(root_path, "generated", "reports")

        # Accept EULA
        eula_path = os.path.join(root_path, "eula.txt")
        if not os.path.exists(eula_path):
            with open(eula_path, "w") as fd:
                fd.write("eula=true\n")

        # Export reports
        if not os.path.exists(reports_path):
            subprocess.check_call(
                ["java", "-cp", jar_name, "net.minecraft.data.Main",
                 "--reports"],
                cwd=root_path)

        # Load data
        return cls.from_json(reports_path)

    @classmethod
    def from_json(cls, reports_path):
        """
        Create a ``LookupRegistry`` from JSON files generated by the official
        server.
        """

        blocks = {}
        registries_temp = {}
        registries = {}

        blocks_path = os.path.join(reports_path, "blocks.json")
        items_path = os.path.join(reports_path, "items.json")
        registries_path = os.path.join(reports_path, "registries.json")

        with open(blocks_path) as fd:
            for name, obj in json.load(fd).items():
                for state in obj['states']:
                    properties = state.get("properties", {})
                    properties['name'] = name
                    blocks[state['id']] = properties

        if os.path.exists(items_path):
            with open(items_path) as fd:
                registries_temp['minecraft:item'] = {'entries': json.load(fd)}

        if os.path.exists(registries_path):
            with open(registries_path) as fd:
                registries_temp.update(json.load(fd))

        for registry_name, registry in registries_temp.items():
            registries[registry_name] = {}
            for name, obj in registry['entries'].items():
                registries[registry_name][obj['protocol_id']] = name

        return cls(blocks, registries)


from quarry.types.buffer.v1_7 import Buffer1_7
from quarry.types.chunk import PackedArray, BlockArray


class Buffer1_9(Buffer1_7):

    # Chunk section -----------------------------------------------------------

    @classmethod
    def pack_chunk(cls, sections):
        data = b""
        for section in sections:
            if section and not section[0].is_empty():
                data += cls.pack_chunk_section(*section)
        return data

    @classmethod
    def pack_chunk_bitmask(cls, sections):
        bitmask = 0
        for i, section in enumerate(sections):
            if section and not section[0].is_empty():
                bitmask |= 1 << i
        return cls.pack_varint(bitmask)

    @classmethod
    def pack_chunk_section(cls, blocks, block_lights, sky_lights=None):
        """
        Packs a chunk section. The supplied arguments should be instances of
        ``BlockArray`` and ``LightArray`` from ``quarry.types.chunk``.
        """

        out = cls.pack('B', blocks.storage.value_width)
        out += cls.pack_chunk_section_palette(blocks.palette)
        out += cls.pack_chunk_section_array(blocks.to_bytes())
        out += block_lights.to_bytes()
        if sky_lights:
            out += sky_lights.to_bytes()
        return out

    @classmethod
    def pack_chunk_section_palette(cls, palette):
        return cls.pack_varint(len(palette)) + b"".join(
            cls.pack_varint(x) for x in palette)

    @classmethod
    def pack_chunk_section_array(cls, data):
        return cls.pack_varint(len(data) // 8) + data

    def unpack_chunk(self, bitmask, overworld=True):
        sections = []
        for idx in range(16):
            if bitmask & (1 << idx):
                section = self.unpack_chunk_section(overworld)
            else:
                section = None
            sections.append(section)
        return sections

    def unpack_chunk_section(self, overworld=True):
        """
        Unpacks a chunk section. Returns a 3-tuple of
        ``(blocks, block_lights, sky_lights)``, where *sky_lights* is ``None``
        when *overworld* is ``False``. The returned values are sequences of
        length 4096 (16x16x16).
        """

        value_width = self.unpack('B')
        palette = self.unpack_chunk_section_palette(value_width)
        array = self.unpack_chunk_section_array(value_width)
        blocks = BlockArray.from_bytes(
            bytes=array,
            palette=palette,
            registry=self.registry,
            non_air=None,
            value_width=value_width)
        block_lights = PackedArray.from_light_bytes(self.read(2048))
        if overworld:
            sky_lights = PackedArray.from_light_bytes(self.read(2048))
        else:
            sky_lights = None

        return blocks, block_lights, sky_lights

    def unpack_chunk_section_palette(self, value_width):
        return [self.unpack_varint() for _ in range(self.unpack_varint())]

    def unpack_chunk_section_array(self, value_width):
        return self.read(self.unpack_varint() * 8)

    # Entity metadata ---------------------------------------------------------

    @classmethod
    def pack_entity_metadata(cls, metadata):
        """
        Packs entity metadata.
        """

        pack_position = lambda pos: cls.pack_position(*pos)
        out = b""
        for ty_key, val in metadata.items():
            ty, key = ty_key
            out += cls.pack('BB', key, ty)
            if   ty == 0:  out += cls.pack('b', val)
            elif ty == 1:  out += cls.pack_varint(val)
            elif ty == 2:  out += cls.pack('f', val)
            elif ty == 3:  out += cls.pack_string(val)
            elif ty == 4:  out += cls.pack_chat(val)
            elif ty == 5:  out += cls.pack_slot(**val)
            elif ty == 6:  out += cls.pack('?', val)
            elif ty == 7:  out += cls.pack_rotation(*val)
            elif ty == 8:  out += cls.pack_position(*val)
            elif ty == 9:  out += cls.pack_optional(pack_position, val)
            elif ty == 10: out += cls.pack_direction(val)
            elif ty == 11: out += cls.pack_optional(cls.pack_uuid, val)
            elif ty == 12: out += cls.pack_block(val)
            elif ty == 13: out += cls.pack_nbt(val)
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
        out += cls.pack('B', 255)
        return out

    def unpack_entity_metadata(self):
        """
        Unpacks entity metadata.
        """

        metadata = {}
        while True:
            key = self.unpack('B')
            if key == 255:
                return metadata
            ty = self.unpack('B')
            if   ty == 0:  val = self.unpack('b')
            elif ty == 1:  val = self.unpack_varint()
            elif ty == 2:  val = self.unpack('f')
            elif ty == 3:  val = self.unpack_string()
            elif ty == 4:  val = self.unpack_chat()
            elif ty == 5:  val = self.unpack_slot()
            elif ty == 6:  val = self.unpack('?')
            elif ty == 7:  val = self.unpack_rotation()
            elif ty == 8:  val = self.unpack_position()
            elif ty == 9:  val = self.unpack_optional(self.unpack_position)
            elif ty == 10: val = self.unpack_direction()
            elif ty == 11: val = self.unpack_optional(self.unpack_uuid)
            elif ty == 12: val = self.unpack_block()
            elif ty == 13: val = self.unpack_nbt()
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
            metadata[ty, key] = val



from collections.abc import Sequence, MutableSequence
from bitstring import BitArray, Bits
import math


def get_width(length, full_width):
    """
    Returns the number of bits used by Minecraft to represent indices into a
    list of the given length.
    """

    width = int(math.ceil(math.log(length, 2)))
    if width < 4:
        return 4
    elif width > 8:
        return full_width
    else:
        return width


class PackedArray(Sequence):
    """
    This class provides support for an array where values are tightly packed
    into a number of bits (such as 4 bits for light or 9 bits for height).

    All operations associated with fixed-size mutable sequences are supported,
    such as slicing.

    You may need to adjust the *length* and *value_width* of packed arrays from
    NBT data, where these fields are not conveyed.

    Several constructors are available for specific uses of packed arrays:

    - Light data used 4-bit values and 8-bit sectors
    - Height data uses 9-bit values and 64-bit sectors
    - Block data uses 64-bit sectors
    """

    #: The ``bitstring.BitArray`` object used for storage.
    storage = None

    #: The number of entries in the array
    length = None

    #: The width in bits of sectors. Used in (de)serialization.
    sector_width = None

    #: The width in bits of values.
    value_width = None


    def __repr__(self):
        return "<PackedArray length=%d sector_width=%d value_width=%d>" \
               % (self.length,
                  self.sector_width,
                  self.value_width)

    # Constructors ------------------------------------------------------------

    def __init__(self, storage, length, sector_width, value_width):
        self.storage = storage
        self.length = length
        self.sector_width = sector_width
        self.value_width = value_width

    @classmethod
    def empty(cls, length, sector_width, value_width):
        """
        Creates an empty array.
        """

        obj = cls(BitArray(), length, sector_width, value_width)
        obj.purge()
        return obj

    @classmethod
    def empty_light(cls):
        """
        Creates an empty array suitable for storing light data.
        """

        return cls.empty(4096, 8, 4)

    @classmethod
    def empty_block(cls):
        """
        Creates an empty array suitable for storing block data.
        """

        return cls.empty(4096, 64, 4)

    @classmethod
    def empty_height(cls):
        """
        Creates an empty array suitable for storing height data.
        """

        return cls.empty(256, 64, 9)

    @classmethod
    def from_bytes(cls, bytes, length, sector_width, value_width):
        """
        Deserialize a packed array from the given bytes.
        """

        storage = BitArray(bytes=bytes)
        return cls(storage, length, sector_width, value_width)


    @classmethod
    def from_light_bytes(cls, bytes):
        """
        Deserialize a packed array from the given light data bytes.
        """

        return cls.from_bytes(bytes, 4096, 8, 4)

    @classmethod
    def from_block_bytes(cls, bytes, value_width):
        """
        Deserialize a packed array from the given block data bytes.
        """

        return cls.from_bytes(bytes, 4096, 64, value_width)

    @classmethod
    def from_height_bytes(cls, bytes):
        """
        Deserialize a packed array from the given height data bytes.
        """

        return cls.from_bytes(bytes, 256, 64, 9)

    # Instance methods --------------------------------------------------------

    def to_bytes(self):
        """
        Serialize this packed array to bytes.
        """

        return self.storage.bytes

    def purge(self):
        """
        Initializes the storage.

        You should not need to call this method.
        """

        values_per_sector = self.sector_width // self.value_width
        sector_count = 1 + (self.length - 1) // values_per_sector
        self.storage.clear()
        self.storage.append(BitArray(self.sector_width * sector_count))

    def pos(self, idx):
        """
        Returns the bit position of the value at the given index.

        You should not need to call this method.
        """

        sector, value = divmod(idx, self.sector_width // self.value_width)
        pos = (1 + sector) * self.sector_width - \
              (1 + value ) * self.value_width
        return pos, pos + self.value_width

    def is_empty(self):
        """
        Returns true if this packed array is entirely zeros.
        """

        return not self.storage.any(True)

    # Sequence methods --------------------------------------------------------

    def __len__(self):
        return self.length

    def __iter__(self):
        for i in range(self.length):
            yield self.storage._slice(*self.pos(i)).uint

    def __getitem__(self, item):
        if isinstance(item, slice):
            return [self.storage._slice(*self.pos(idx)).uint
                    for idx in range(*item.indices(len(self)))]
        else:
            if not 0 <= item < len(self):
                raise IndexError(item)
            return self.storage._slice(*self.pos(item)).uint

    def __setitem__(self, item, value):
        if isinstance(item, slice):
            for idx, value in zip(range(*item.indices(len(self))), value):
                self.storage._overwrite(
                    Bits(uint=value, length=self.value_width),
                    self.pos(idx)[0])
        else:
            self.storage._overwrite(
                Bits(uint=value, length=self.value_width),
                self.pos(item)[0])


class BlockArray(Sequence):
    """
    This class provides support for block arrays. It wraps a
    :class:`PackedArray` object and implements block encoding/decoding,
    palettes, and counting of non-air blocks for lighting purposes. It stores
    precisely 4096 (16x16x16) values.

    All operations associated with fixed-size mutable sequences are supported,
    such as slicing.

    A palette is used when there are fewer than 256 unique values; the value
    width varies from 4 to 8 bits depending on the size of the palette, and is
    automatically adjusted upwards as necessary. Use :meth:`~BlockArray.repack`
    to reclaim space by eliminating unused entries.

    When 256 or more unique values are present, the palette is unused and
    values are stored directly.
    """

    #: The :class:`PackedArray` object used for storage.
    storage = None

    #: List of encoded block values. Empty when palette is not used.
    palette = None

    #: The `Registry` object used to encode/decode blocks
    registry = None

    #: The number of non-air blocks
    non_air = None

    def __repr__(self):
        return "<BlockArray palette=%d storage=%r>" \
               % (len(self.palette), self.storage)

    # Constructors ------------------------------------------------------------

    def __init__(self, storage, palette, registry, non_air=-1):
        self.storage = storage
        self.palette = palette
        self.registry = registry
        self._non_air = non_air

    @classmethod
    def empty(cls, registry, non_air=-1):
        """
        Creates an empty block array.
        """

        storage = PackedArray.empty(4096, 64, 4)
        palette = [0]
        return cls(storage, palette, registry, non_air)

    @classmethod
    def from_bytes(cls, bytes, value_width, registry, palette, non_air=-1):
        """
        Deserialize a block array from the given bytes.
        """
        storage = PackedArray.from_block_bytes(bytes, value_width)
        return cls(storage, palette, registry, non_air)

    @classmethod
    def from_nbt(cls, section, registry, non_air=-1):
        """
        Creates a block array that uses the given NBT section tag as storage
        for block data and the palette. Minecraft 1.13+ only.
        """

        nbt_palette = section.value['Palette']
        if isinstance(nbt_palette.value, _NBTPaletteProxy):
            proxy = nbt_palette.value
        else:
            proxy = _NBTPaletteProxy(registry)
            for entry in nbt_palette.value:
                proxy.append(entry)
            nbt_palette.value = proxy

        storage = section.value["BlockStates"].value
        palette = proxy.palette
        storage.length = 4096
        storage.value_width = get_width(len(proxy), registry.max_bits)
        return cls(storage, palette, registry, non_air)

    # Instance methods --------------------------------------------------------

    def to_bytes(self):
        """
        Serialize this block array to bytes.
        """

        return self.storage.to_bytes()

    def is_empty(self):
        """
        Returns true if this block array is entirely air.
        """

        if self.palette == [0]:
            return True
        else:
            return self.non_air == 0

    @property
    def non_air(self):
        if self._non_air == -1:
            self._non_air = [
                self.registry.is_air_block(obj) for obj in self].count(False)
        return self._non_air

    def repack(self, reserve=None):
        """
        Re-packs internal data to use the smallest possible bits-per-block by
        eliminating unused palette entries. This operation is slow as it walks
        all blocks to determine the new palette.
        """

        # If no reserve is given, we re-compute the palette by walking blocks
        if reserve is None:
            palette = sorted(set(self))
            palette_len = len(palette)

        # Otherwise we just ensure we have enough space to store new entries.
        elif self.palette:
            palette = self.palette[:]
            palette_len = len(palette) + reserve

        # Reserving space in an unpaletted array is a no-op.
        else:
            return

        # Compute new value width
        value_width = get_width(palette_len, self.registry.max_bits)

        # Exit if there's no change in value width needed
        if value_width == self.storage.value_width:
            return

        # Switch to unpaletted operation if necessary
        if value_width > 8:
            palette = []

        # Save contents
        values = self[:]

        # Update internals
        self.storage.value_width = value_width
        self.storage.purge()
        self.palette[:] = palette

        # Load contents
        self[:] = values

    # Sequence methods --------------------------------------------------------

    def __len__(self):
        return 4096

    def __getitem__(self, item):
        if isinstance(item, slice):
            values = []
            for value in self.storage[item.start:item.stop:item.step]:
                if self.palette:
                    value = self.palette[value]
                value = self.registry.decode_block(value)
                values.append(value)
            return values
        else:
            value = self.storage[item]
            if self.palette:
                value = self.palette[value]
            value = self.registry.decode_block(value)
            return value

    def __setitem__(self, item, value):
        if isinstance(item, slice):
            for idx in range(*item.indices(4096)):
                self[idx] = value[idx]
            return

        if self._non_air != -1:
            self._non_air += int(self.registry.is_air_block(self[item])) - \
                             int(self.registry.is_air_block(value))

        value = self.registry.encode_block(value)

        if self.palette:
            try:
                value = self.palette.index(value)
            except ValueError:
                self.repack(reserve=1)

                if self.palette:
                    self.palette.append(value)
                    value = len(self.palette) - 1

        self.storage[item] = value

    def __iter__(self):
        for value in self.storage:
            if self.palette:
                value = self.palette[value]
            value = self.registry.decode_block(value)
            yield value

    def __contains__(self, value):
        if self.palette:
            if self.registry.encode_block(value) not in self.palette:
                return False
        return super(BlockArray, self).__contains__(value)

    def index(self, value, start=0, stop=None):
        if self.palette:
            if self.registry.encode_block(value) not in self.palette:
                raise ValueError
        return super(BlockArray, self).index(value, start, stop)

    def count(self, value):
        if self.palette:
            if self.registry.encode_block(value) not in self.palette:
                return 0
        return super(BlockArray, self).count(value)


class _NBTPaletteProxy(MutableSequence):
    def __init__(self, registry):
        self.registry = registry
        self.palette = []

    def insert(self, idx, value):
        # FIXME: NBT chunk sections are *always* paletted, and so the format
        # diverges for palettes longer than 255 entries.
        if len(self.palette) >= 255:
            raise ValueError("Can't add more than 255 entries to NBT palette "
                             "proxy.")
        self.palette.insert(idx, None)
        self[idx] = value

    def __len__(self):
        return len(self.palette)

    def __delitem__(self, idx):
        del self.palette[idx]

    def __getitem__(self, idx):
        from quarry.types import nbt

        block = self.registry.decode_block(self.palette[idx])
        entry = nbt.TagCompound({'Name': nbt.TagString(block['name'])})
        if len(block) > 1:
            entry.value['Properties'] = nbt.TagCompound({
                key: nbt.TagString(value)
                for key, value in block.items()
                if key != "name"})

        return entry

    def __setitem__(self, idx, tag):
        block = {'name': tag.value['Name'].value}
        properties = tag.value.get('Properties')
        if properties:
            block.update(properties.to_obj())

        self.palette[idx] = self.registry.encode_block(block)



from quarry.types.chunk import BlockArray
from quarry.types.buffer.v1_13_2 import Buffer1_13_2

poses = ('standing', 'fall_flying', 'sleeping', 'swimming', 'spin_attack',
         'sneaking', 'dying')
smelt_types = ('minecraft:smelting', 'minecraft:blasting',
               'minecraft:smoking', 'minecraft:campfire_cooking')


class Buffer1_14(Buffer1_13_2):

    # Chunk section -----------------------------------------------------------

    @classmethod
    def pack_chunk_section(cls, blocks, block_lights=None, sky_lights=None):
        """
        Packs a chunk section. The supplied argument should be an instance of
        ``quarry.types.chunk.BlockArray``.
        """

        out = cls.pack('HB', blocks.non_air, blocks.storage.value_width)
        out += cls.pack_chunk_section_palette(blocks.palette)
        out += cls.pack_chunk_section_array(blocks.to_bytes())
        return out

    def unpack_chunk_section(self, overworld=True):
        """
        Unpacks a chunk section. Returns a sequence of length 4096 (16x16x16).
        """

        non_air, value_width = self.unpack('HB')
        palette = self.unpack_chunk_section_palette(value_width)
        array = self.unpack_chunk_section_array(value_width)
        return BlockArray.from_bytes(
            bytes=array,
            palette=palette,
            registry=self.registry,
            non_air=non_air,
            value_width=value_width), None, None

    # Position ----------------------------------------------------------------

    @classmethod
    def pack_position(cls, x, y, z):
        """
        Packs a Position.
        """

        def pack_twos_comp(bits, number):
            if number < 0:
                number = number + (1 << bits)
            return number

        return cls.pack('Q', sum((
            pack_twos_comp(26, x) << 38,
            pack_twos_comp(26, z) << 12,
            pack_twos_comp(12, y))))

    def unpack_position(self):
        """
        Unpacks a position.
        """

        def unpack_twos_comp(bits, number):
            if (number & (1 << (bits - 1))) != 0:
                number = number - (1 << bits)
            return number

        number = self.unpack('Q')
        x = unpack_twos_comp(26, (number >> 38))
        z = unpack_twos_comp(26, (number >> 12 & 0x3FFFFFF))
        y = unpack_twos_comp(12, (number & 0xFFF))
        return x, y, z


    # Entity metadata ---------------------------------------------------------

    @classmethod
    def pack_entity_metadata(cls, metadata):
        """
        Packs entity metadata.
        """

        pack_position = lambda pos: cls.pack_position(*pos)
        out = b""
        for ty_key, val in metadata.items():
            ty, key = ty_key
            out += cls.pack('BB', key, ty)
            if   ty == 0:  out += cls.pack('b', val)
            elif ty == 1:  out += cls.pack_varint(val)
            elif ty == 2:  out += cls.pack('f', val)
            elif ty == 3:  out += cls.pack_string(val)
            elif ty == 4:  out += cls.pack_chat(val)
            elif ty == 5:  out += cls.pack_optional(cls.pack_chat, val)
            elif ty == 6:  out += cls.pack_slot(**val)
            elif ty == 7:  out += cls.pack('?', val)
            elif ty == 8:  out += cls.pack_rotation(*val)
            elif ty == 9:  out += cls.pack_position(*val)
            elif ty == 10: out += cls.pack_optional(pack_position, val)
            elif ty == 11: out += cls.pack_direction(val)
            elif ty == 12: out += cls.pack_optional(cls.pack_uuid, val)
            elif ty == 13: out += cls.pack_block(val)
            elif ty == 14: out += cls.pack_nbt(val)
            elif ty == 15: out += cls.pack_particle(*val)
            elif ty == 16: out += cls.pack_villager(*val)
            elif ty == 17: out += cls.pack_optional_varint(val)
            elif ty == 18: out += cls.pack_pose(val)
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
        out += cls.pack('B', 255)
        return out

    def unpack_entity_metadata(self):
        """
        Unpacks entity metadata.
        """

        metadata = {}
        while True:
            key = self.unpack('B')
            if key == 255:
                return metadata
            ty = self.unpack('B')
            if   ty == 0:  val = self.unpack('b')
            elif ty == 1:  val = self.unpack_varint()
            elif ty == 2:  val = self.unpack('f')
            elif ty == 3:  val = self.unpack_string()
            elif ty == 4:  val = self.unpack_chat()
            elif ty == 5:  val = self.unpack_optional(self.unpack_chat)
            elif ty == 6:  val = self.unpack_slot()
            elif ty == 7:  val = self.unpack('?')
            elif ty == 8:  val = self.unpack_rotation()
            elif ty == 9:  val = self.unpack_position()
            elif ty == 10: val = self.unpack_optional(self.unpack_position)
            elif ty == 11: val = self.unpack_direction()
            elif ty == 12: val = self.unpack_optional(self.unpack_uuid)
            elif ty == 13: val = self.unpack_block()
            elif ty == 14: val = self.unpack_nbt()
            elif ty == 15: val = self.unpack_particle()
            elif ty == 16: val = self.unpack_villager()
            elif ty == 17: val = self.unpack_optional_varint()
            elif ty == 18: val = self.unpack_pose()
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
            metadata[ty, key] = val

    # Particle ----------------------------------------------------------------

    @classmethod
    def pack_particle(cls, kind, data=None):
        """
        Packs a particle.
        """

        id = cls.registry.encode('minecraft:particle_type', kind)
        return super(Buffer1_14, cls).pack_particle(id, data)


    def unpack_particle(self):
        """
        Unpacks a particle. Returns an ``(kind, data)`` pair.
        """

        id, data = super(Buffer1_14, self).unpack_particle()
        kind = self.registry.decode('minecraft:particle_type', id)
        return kind, data


    # Villager data -----------------------------------------------------------

    @classmethod
    def pack_villager(cls, kind, profession, level):
        """
        Packs villager data.
        """

        kind = cls.registry.encode('minecraft:villager_type', kind)
        profession = cls.registry.encode('minecraft:villager_profession', profession)
        return cls.pack_varint(kind) + \
               cls.pack_varint(profession) + \
               cls.pack_varint(level)

    def unpack_villager(self):
        """
        Unpacks villager data.
        """
        kind = self.registry.decode(
            'minecraft:villager_type', self.unpack_varint())
        profession = self.registry.decode(
            'minecraft:villager_profession', self.unpack_varint())
        level = self.unpack_varint()
        return kind, profession, level


    # Optional varint ---------------------------------------------------------

    @classmethod
    def pack_optional_varint(cls, number):
        """
        Packs an optional varint.
        """

        return cls.pack_varint(0 if number is None else number + 1)

    def unpack_optional_varint(self):
        """
        Unpacks an optional varint.
        """

        val = self.unpack_varint()
        if val == 0:
            return None
        else:
            return val - 1


    # Pose --------------------------------------------------------------------

    @classmethod
    def pack_pose(cls, pose):
        """
        Packs a pose.
        """

        return cls.pack_varint(poses.index(pose))

    def unpack_pose(self):
        """
        Unpacks a pose.
        """

        return poses[self.unpack_varint()]

    # Recipes -----------------------------------------------------------------

    def unpack_recipe(self):
        """
        Unpacks a crafting recipe.
        """
        recipe = {}
        recipe['type'] = self.unpack_string()
        recipe['name'] = self.unpack_string()

        if recipe['type'] == 'minecraft:crafting_shapeless':
            recipe['group'] = self.unpack_string()
            recipe['ingredients'] = [
                self.unpack_ingredient() for _ in range(self.unpack_varint())]
            recipe['result'] = self.unpack_slot()

        elif recipe['type'] == 'minecraft:crafting_shaped':
            recipe['width'] = self.unpack_varint()
            recipe['height'] = self.unpack_varint()
            recipe['group'] = self.unpack_string()
            recipe['ingredients'] = [
                self.unpack_ingredient() for _ in range(recipe['width'] *
                                                    recipe['height'])]
            recipe['result'] = self.unpack_slot()
        elif recipe['type'] in smelt_types:
            recipe['group'] = self.unpack_string()
            recipe['ingredient'] = self.unpack_ingredient()
            recipe['result'] = self.unpack_slot()
            recipe['experience'] = self.unpack('f')
            recipe['cooking_time'] = self.unpack_varint()

        return recipe

    @classmethod
    def pack_recipe(cls, name, type, **recipe):
        """
        Packs a crafting recipe.
        """
        data = cls.pack_string(type) + cls.pack_string(name)

        if type == 'minecraft:crafting_shapeless':
            data += cls.pack_string(recipe['group'])
            data += cls.pack_varint(len(recipe['ingredients']))
            for ingredient in recipe['ingredients']:
                data += cls.pack_ingredient(ingredient)
            data += cls.pack_slot(**recipe['result'])

        elif type == 'minecraft:crafting_shaped':
            data += cls.pack_varint(recipe['width'])
            data += cls.pack_varint(recipe['height'])
            data += cls.pack_string(recipe['group'])
            for ingredient in recipe['ingredients']:
                data += cls.pack_ingredient(ingredient)
            data += cls.pack_slot(**recipe['result'])

        elif type in smelt_types:
            data += cls.pack_string(recipe['group'])
            data += cls.pack_ingredient(recipe['ingredient'])
            data += cls.pack_slot(**recipe['result'])
            data += cls.pack('f', recipe['experience'])
            data += cls.pack_varint(recipe['cooking_time'])

        return data



import logging

from quarry.net.protocol import PacketDispatcher
from quarry.net.server import ServerFactory, ServerProtocol
from quarry.net.client import ClientFactory, ClientProtocol
from quarry.net.auth import OfflineProfile


def _enable_forwarding(endpoint):
    """
    Patches the given endpoint's ``packet_received()`` method to pass packets
    through the bridge.
    """
    def packet_received(buff, name):
        endpoint.bridge.packet_received(
            buff,
            endpoint.recv_direction,
            name)
    endpoint._packet_received = endpoint.packet_received
    endpoint.packet_received = packet_received


def _disable_forwarding(endpoint):
    """
    Patches the given endpoint's ``packet_received()`` method to restore
    handling of packets within the endpoint.
    """
    endpoint.packet_received = endpoint._packet_received


def _enable_fast_forwarding(endpoint1, endpoint2):
    """
    Patches the first given endpoint's ``data_received()`` method to network
    data directly to the second endpoint, without any packet decoding.
    """
    if len(endpoint1.recv_buff) > 0:
        endpoint2.transport.write(
            endpoint2.cipher.encrypt(
                endpoint1.recv_buff.read()))

    def data_received(data):
        endpoint2.transport.write(
            endpoint2.cipher.encrypt(
                endpoint1.cipher.decrypt(data)))

    endpoint1.data_received = data_received


class Upstream(ClientProtocol):
    def setup(self):
        self.bridge = self.factory.bridge
        self.bridge.upstream = self

    def player_joined(self):
        self.bridge.upstream_ready()

    def connection_lost(self, reason=None):
        ClientProtocol.connection_lost(self, reason)
        self.bridge.upstream_disconnected()


class UpstreamFactory(ClientFactory):
    protocol = Upstream
    bridge = None


class Bridge(PacketDispatcher):
    """
    This class exchanges packets between the upstream and downstream.
    """

    upstream_factory_class = UpstreamFactory
    log_level = logging.INFO

    logger = None
    buff_type = None

    downstream_factory = None
    downstream = None

    upstream_profile = None
    upstream_factory = None
    upstream = None

    def __init__(self, downstream_factory, downstream):
        self.downstream_factory = downstream_factory
        self.downstream = downstream

        self.buff_type = self.downstream.buff_type

        self.logger = logging.getLogger("%s{%s}" % (
            self.__class__.__name__,
            self.downstream.remote_addr.host))
        self.logger.setLevel(self.log_level)

    def make_profile(self):
        """
        Returns the profile to use for the upstream connection. By default, use
        an offline profile with the same display name as the remote client.
        """
        return OfflineProfile(self.downstream.display_name)

    def connect(self):
        """
        Connect to the remote server.
        """

        self.upstream_profile = self.make_profile()
        self.upstream_factory = self.upstream_factory_class(
            self.upstream_profile)
        self.upstream_factory.bridge = self
        self.upstream_factory.force_protocol_version = \
            self.downstream.protocol_version
        self.upstream_factory.connect(
            self.connect_host,
            self.connect_port)

    # Connections -------------------------------------------------------------

    def downstream_ready(self):
        """
        Called when the downstream is waiting for forwarding to begin.
        By default, this method begins a connection to the remote server.
        """

        self.logger.debug("Downstream ready")

        # Connect to the server the client is requesting
        if self.downstream_factory.connect_host is None:
            self.connect_host = self.downstream.connect_host
            self.connect_port = self.downstream.connect_port
        else:
            self.connect_host = self.downstream_factory.connect_host
            self.connect_port = self.downstream_factory.connect_port

        self.connect()

    def upstream_ready(self):
        """
        Called when the upstream is waiting for forwarding to begin. By
        default, enables forwarding.
        """
        self.logger.debug("Upstream ready")
        self.enable_forwarding()

    def downstream_disconnected(self):
        """
        Called when the connection to the remote client is closed.
        """
        if self.upstream:
            self.upstream.close()

    def upstream_disconnected(self):
        """
        Called when the connection to the remote server is closed.
        """
        self.downstream.close("Lost connection to server.")

    # Pass through ------------------------------------------------------------

    def enable_forwarding(self):
        """
        Enables forwarding. Packet handlers in the ``Upstream`` and
        ``Downstream`` cease to be called, and all packets are routed via the
        ``Bridge``. This method is called by ``upstream_ready()`` by default.
        """

        _enable_forwarding(self.downstream)
        _enable_forwarding(self.upstream)
        self.logger.debug("Forwarding enabled")

    def disable_forwarding(self):
        """
        Disable forwarding. Packet handlers in the ``Bridge`` cease to be
        called, and packets are routed via the ``Upstream`` and ``Downstream``.
        This method is not called by default.
        """

        _disable_forwarding(self.downstream)
        _disable_forwarding(self.upstream)
        self.logger.debug("Forwarding disabled")

    def enable_fast_forwarding(self):
        """
        Enables fast forwarding. Quarry passes network data between endpoints
        without decoding packets, and therefore all packet handlers cease to be
        called. Both parts of the proxy must be operating at the same
        compression threshold. This method is not called by default.
        """
        if self.downstream.compression_threshold != \
                self.upstream.compression_threshold:
            raise Exception(
                "Cannot enable fast forwarding as compression differs. "
                "downstream: %s, upstream: %s" % (
                    self.downstream.compression_threshold,
                    self.upstream.compression_threshold))

        _enable_fast_forwarding(self.downstream, self.upstream)
        _enable_fast_forwarding(self.upstream, self.downstream)
        self.logger.debug("Fast forwarding enabled")

    # Packet handling ---------------------------------------------------------

    def packet_received(self, buff, direction, name):
        """
        Called when a packet is received a remote. Usually this method
        dispatches the packet to a method named
        ``packet_<direction>_<packet name>``, or calls :meth:`packet_unhandled`
        if no such methods exists. You might want to override this to implement
        your own dispatch logic or logging.
        """

        dispatched = self.dispatch((direction, name), buff)

        if not dispatched:
            self.packet_unhandled(buff, direction, name)

    def packet_unhandled(self, buff, direction, name):
        """
        Called when a packet is received that is not hooked. The default
        implementation forwards the packet.
        """
        if direction == "downstream":
            self.downstream.send_packet(name, buff.read())
        elif direction == "upstream":
            self.upstream.send_packet(name, buff.read())

    def packet_downstream_set_compression(self, buff):
        self.upstream.set_compression(buff.unpack_varint())


class Downstream(ServerProtocol):
    bridge = None

    def setup(self):
        self.bridge = self.factory.bridge_class(self.factory, self)

    def player_joined(self):
        ServerProtocol.player_joined(self)
        self.bridge.downstream_ready()

    def connection_lost(self, reason=None):
        ServerProtocol.connection_lost(self, reason)
        self.bridge.downstream_disconnected()


class DownstreamFactory(ServerFactory):
    protocol = Downstream
    connect_host = None
    connect_port = None
    bridge_class = Bridge



import argparse
import json
import os
import sys

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from twisted.internet import defer
from quarry.net import http
from quarry.types.uuid import UUID

if sys.version_info[0] == 2:
    from urllib import urlencode
else:
    from urllib.parse import urlencode

SESSION_SERVER = b"https://sessionserver.mojang.com/session/minecraft/"


class ProfileException(http.HTTPException):
    pass


class AuthException(http.HTTPException):
    pass


class OfflineProfile(object):
    online = False

    def __init__(self, display_name="quarry"):
        self.display_name = display_name

    @classmethod
    def from_display_name(cls, display_name):
        return cls(display_name)


class Profile(object):
    online = True
    timeout = 30

    def __init__(self, client_token, access_token, display_name, uuid):
        self.client_token = client_token
        self.access_token = access_token
        self.display_name = display_name
        self.uuid = uuid

    def join(self, digest, refresh=True):
        d1 = http.request(
            url=SESSION_SERVER + b"join",
            timeout=self.timeout,
            err_type=AuthException,
            data={"accessToken": self.access_token,
                  "selectedProfile": self.uuid.to_hex(with_dashes=False),
                  "serverId": digest})

        if not refresh:
            return d1
        else:
            d0 = defer.Deferred()

            def _errback(err):
                self.refresh().chainDeferred(
                    self.join(digest, refresh=False).chainDeferred(d0))
            d1.addCallbacks(d0.callback, _errback)
            return d0

    def validate(self):
        d0 = defer.Deferred()

        def _callback(data):
            d0.callback(self)

        d1 = self._request(b"validate", accessToken=self.access_token)
        d1.addCallbacks(_callback, d0.errback)
        return d0

    def refresh(self):
        d0 = defer.Deferred()

        def _callback(data):
            d0.callback(self)

        d1 = self._request(
            b"refresh",
            clientToken=self.client_token,
            accessToken=self.access_token)
        d1.addCallbacks(_callback, d0.errback)
        return d0

    def to_file(self, profiles_path=None):
        if not profiles_path:
            profiles_path = self._get_profiles_path()

        with open(profiles_path, "w") as fd:
            json.dump({
                "selectedUser": self.uuid.to_hex(False),
                "clientToken": self.client_token,
                "authenticationDatabase": {
                    self.uuid.to_hex(False): {
                        "displayName": self.display_name,
                        "accessToken": self.access_token,
                        "uuid": self.uuid.to_hex(True)}}}, fd)

    @classmethod
    def from_credentials(cls, email, password):
        d0 = defer.Deferred()

        def _callback(data):
            d0.callback(cls._from_response(data))

        def _errback(err):
            d0.errback(err)

        agent = {
            "name": "Minecraft",
            "version": 1
        }

        client_token = "foo"  # TODO

        d1 = cls._request(
            b"authenticate",
            username=email,
            password=password,
            agent=agent,
            clientToken=client_token)
        d1.addCallbacks(_callback, _errback)

        return d0

    @classmethod
    def from_token(cls, client_token, access_token, display_name, uuid):
        obj = cls(client_token, access_token,
                  display_name, UUID.from_hex(uuid))
        return obj.validate()

    @classmethod
    def from_file(cls, display_name=None, uuid=None, profiles_path=None):
        if profiles_path is None:
            profiles_path = cls._get_profiles_path()

        with open(profiles_path) as fd:
            data = json.load(fd)

        client_token = data["clientToken"]

        if display_name is None and uuid is None:
            uuid = data["selectedUser"]["profile"]

        for store in data["authenticationDatabase"].values():
            access_token = store["accessToken"]

            for p_uuid, p_data in store["profiles"].items():
                p_display_name = p_data["displayName"]
                if uuid and uuid != p_uuid:
                    continue
                if display_name and display_name != p_display_name:
                    continue
                return cls.from_token(client_token, access_token,
                                      p_display_name, p_uuid)

    @classmethod
    def _from_response(cls, response):
        return cls(
            response['clientToken'],
            response['accessToken'],
            response['selectedProfile']['name'],
            UUID.from_hex(response['selectedProfile']['id']))

    @classmethod
    def _request(cls, endpoint, **data):
        return http.request(
            url=b"https://authserver.mojang.com/" + endpoint,
            timeout=cls.timeout,
            err_type=ProfileException,
            data=data)

    @classmethod
    def _get_profiles_path(cls):
        dot_minecraft = ".minecraft"
        if sys.platform == 'win32':
            app_data = os.environ['APPDATA']
        elif sys.platform == 'darwin':
            app_data = os.path.expanduser("~/Library/Application Support")
            dot_minecraft = "minecraft"
        else:
            app_data = os.path.expanduser("~")
        return os.path.join(
            app_data, dot_minecraft, "launcher_profiles.json")


class ProfileCLI(object):
    @classmethod
    def make_parser(cls, parser=None):
        if parser is None:
            parser = argparse.ArgumentParser()
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            "--auth",
            metavar="EMAIL:PASSWORD",
            help="Sets the Mojang account email address and password with "
                 "which to log in.")
        group.add_argument(
            "--session-name",
            metavar="DISPLAY_NAME",
            help="Sets the display name to look up in the "
                 "~/.minecraft/launcher_profiles.json file. This is used to "
                 "resume an existing logged-in session from the official "
                 "client.")
        group.add_argument(
            "--offline-name",
            metavar="DISPLAY_NAME",
            help="Sets the offline display name to use. If none of these "
                 "options are given, quarry uses 'quarry' as an offline "
                 "display name.")
        return parser

    @classmethod
    def make_profile(cls, args):
        if args.auth:
            email, password = args.auth.split(":", 1)
            return Profile.from_credentials(email, password)
        if args.session_name:
            return Profile.from_file(args.session_name)
        return defer.succeed(
            OfflineProfile.from_display_name(args.offline_name or "quarry"))


class PlayerPublicKey:
    def __init__(self, expiry: int, key: RSAPublicKey, signature: bytes):
        self.key = key
        self.signature = signature
        self.expiry = expiry


def has_joined(timeout, digest, display_name, remote_host=None):
    data = {"username": display_name, "serverId": digest}
    if remote_host:
        data["ip"] = remote_host

    return http.request(
        url=SESSION_SERVER + b"hasJoined?" + urlencode(data).encode('ascii'),
        timeout=timeout,
        err_type=AuthException,
        expect_content=True)



import json
import string
import struct
import zlib

from quarry.types.buffer import BufferUnderrun
from quarry.types.registry import OpaqueRegistry
from quarry.types.uuid import UUID


directions = ("down", "up", "north", "south", "west", "east")


class Buffer1_7(object):
    buff = b""
    pos = 0
    registry = OpaqueRegistry(13)

    def __init__(self, data=None):
        if data:
            self.buff = data

    def __len__(self):
        return len(self.buff) - self.pos

    def add(self, data):
        """
        Add some bytes to the end of the buffer.
        """

        self.buff += data

    def save(self):
        """
        Saves the buffer contents.
        """

        self.buff = self.buff[self.pos:]
        self.pos = 0

    def restore(self):
        """
        Restores the buffer contents to its state when :meth:`save` was last
        called.
        """

        self.pos = 0

    def discard(self):
        """
        Discards the entire buffer contents.
        """

        self.pos = len(self.buff)

    def read(self, length=None):
        """
        Read *length* bytes from the beginning of the buffer, or all bytes if
        *length* is ``None``
        """

        if length is None:
            data = self.buff[self.pos:]
            self.pos = len(self.buff)
        else:
            if self.pos + length > len(self.buff):
                raise BufferUnderrun()

            data = self.buff[self.pos:self.pos+length]
            self.pos += length

        return data

    def hexdump(self):
        data = self.buff[self.pos:]
        lines = ['']
        bytes_read = 0
        while len(data) > 0:
            data_line, data = data[:16], data[16:]

            l_hex = []
            l_str = []
            for i, c in enumerate(data_line):
                l_hex.append(f"{c:02x}")
                c_str = data_line[i:i + 1]
                l_str.append(c_str if c_str in string.printable else ".")

            l_hex.extend(['  '] * (16 - len(l_hex)))
            l_hex.insert(8, '')

            lines.append(f"{bytes_read:08x}  {' '.join(l_hex)}  |{''.join(l_str)}|")

            bytes_read += len(data_line)

        return "\n    ".join(lines + [f"{bytes_read:08x}"])

    # Basic data types --------------------------------------------------------

    @classmethod
    def pack(cls, fmt, *fields):
        """
        Pack *fields* into a struct. The format accepted is the same as for
        ``struct.pack()``.
        """

        return struct.pack(">"+fmt, *fields)

    def unpack(self, fmt):
        """
        Unpack a struct. The format accepted is the same as for
        ``struct.unpack()``.
        """
        fmt = ">" + fmt
        data = self.read(struct.calcsize(fmt))
        fields = struct.unpack(fmt, data)
        if len(fields) == 1:
            fields = fields[0]
        return fields

    # Array data types --------------------------------------------------------

    @classmethod
    def pack_array(cls, fmt, array):
        """
        Packs *array* into a struct. The format accepted is the same as for
        ``struct.pack()``.
        """
        return struct.pack(">" + fmt * len(array), *array)

    def unpack_array(self, fmt, length):
        """
        Unpack an array struct. The format accepted is the same as for
        ``struct.unpack()``.
        """
        data = self.read(struct.calcsize(">" + fmt) * length)
        return list(struct.unpack(">" + fmt * length, data))

    @classmethod
    def pack_byte_array(cls, data):
        """
        Packs an array of bytes preceded by varint denoting the array length.
        """
        return cls.pack_varint(len(data)) + data

    def unpack_byte_array(self):
        """
        Unpack an array of bytes preceded by varint denoting the array length.
        """
        return self.read(self.unpack_varint())

    # Optional ----------------------------------------------------------------

    @classmethod
    def pack_optional(cls, packer, val):
        """
        Packs a boolean indicating whether *val* is None. If not,
        ``packer(val)`` is appended to the returned string.
        """

        if val is None:
            return cls.pack('?', False)
        else:
            return cls.pack('?', True) + packer(val)

    def unpack_optional(self, unpacker):
        """
        Unpacks a boolean. If it's True, return the value of ``unpacker()``.
        Otherwise return None.
        """
        if self.unpack('?'):
            return unpacker()
        else:
            return None

    # Varint ------------------------------------------------------------------

    @classmethod
    def pack_varint(cls, number, max_bits=32):
        """
        Packs a varint.
        """

        number_min = -1 << (max_bits - 1)
        number_max = +1 << (max_bits - 1)
        if not (number_min <= number < number_max):
            raise ValueError(f"varint does not fit in range: {number_min:d} <= {number:d} < {number_max:d}")

        if number < 0:
            number += 1 << 32

        out = b""
        for i in range(10):
            b = number & 0x7F
            number >>= 7
            out += cls.pack("B", b | (0x80 if number > 0 else 0))
            if number == 0:
                break
        return out

    def unpack_varint(self, max_bits=32):
        """
        Unpacks a varint.
        """

        number = 0
        for i in range(10):
            b = self.unpack("B")
            number |= (b & 0x7F) << 7*i
            if not b & 0x80:
                break

        if number & (1 << 31):
            number -= 1 << 32

        number_min = -1 << (max_bits - 1)
        number_max = +1 << (max_bits - 1)
        if not (number_min <= number < number_max):
            raise ValueError(f"varint does not fit in range: {number_min:d} <= {number:d} < {number_max:d}")

        return number

    # Packet ------------------------------------------------------------------

    @classmethod
    def pack_packet(cls, data, compression_threshold=-1):
        """
        Packs a packet frame. This method handles length-prefixing and
        compression.
        """

        if compression_threshold >= 0:
            # Compress data and prepend uncompressed data length
            if len(data) >= compression_threshold:
                data = cls.pack_varint(len(data)) + zlib.compress(data)
            else:
                data = cls.pack_varint(0) + data

        # Prepend packet length
        return cls.pack_varint(len(data), max_bits=32) + data

    def unpack_packet(self, cls, compression_threshold=-1):
        """
        Unpacks a packet frame. This method handles length-prefixing and
        compression.
        """
        body = self.read(self.unpack_varint(max_bits=32))
        buff = cls(body)
        if compression_threshold >= 0:
            uncompressed_length = buff.unpack_varint()
            if uncompressed_length > 0:
                body = zlib.decompress(buff.read())
                buff = cls(body)

        return buff

    # String ------------------------------------------------------------------

    @classmethod
    def pack_string(cls, text):
        """
        Pack a varint-prefixed utf8 string.
        """

        text = text.encode("utf-8")
        return cls.pack_varint(len(text), max_bits=16) + text

    def unpack_string(self):
        """
        Unpack a varint-prefixed utf8 string.
        """

        length = self.unpack_varint(max_bits=16)
        text = self.read(length).decode("utf-8")
        return text

    # JSON --------------------------------------------------------------------

    @classmethod
    def pack_json(cls, obj):
        """
        Serialize an object to JSON and pack it to a Minecraft string.
        """
        return cls.pack_string(json.dumps(obj))

    def unpack_json(self):
        """
        Unpack a Minecraft string and interpret it as JSON.
        """

        obj = json.loads(self.unpack_string())
        return obj

    # Chat --------------------------------------------------------------------

    @classmethod
    def pack_chat(cls, message):
        """
        Pack a Minecraft chat message.
        """
        from quarry.types import chat
        if not isinstance(message, chat.Message):
            message = chat.Message.from_string(message)
        return message.to_bytes()

    def unpack_chat(self):
        """
        Unpack a Minecraft chat message.
        """
        from quarry.types import chat
        return chat.Message.from_buff(self)

    # UUID --------------------------------------------------------------------

    @classmethod
    def pack_uuid(cls, uuid):
        """
        Packs a UUID.
        """

        return uuid.to_bytes()

    def unpack_uuid(self):
        """
        Unpacks a UUID.
        """

        return UUID.from_bytes(self.read(16))

    # Position ----------------------------------------------------------------

    @classmethod
    def pack_position(cls, x, y, z):
        """
        Packs a Position.
        """

        def pack_twos_comp(bits, number):
            if number < 0:
                number = number + (1 << bits)
            return number

        return cls.pack('Q', sum((
            pack_twos_comp(26, x) << 38,
            pack_twos_comp(12, y) << 26,
            pack_twos_comp(26, z))))

    def unpack_position(self):
        """
        Unpacks a position.
        """

        def unpack_twos_comp(bits, number):
            if (number & (1 << (bits - 1))) != 0:
                number = number - (1 << bits)
            return number

        number = self.unpack('Q')
        x = unpack_twos_comp(26, (number >> 38))
        y = unpack_twos_comp(12, (number >> 26 & 0xFFF))
        z = unpack_twos_comp(26, (number & 0x3FFFFFF))
        return x, y, z

    # Block -------------------------------------------------------------------

    @classmethod
    def pack_block(cls, block, packer=None):
        """
        Packs a block.
        """
        if packer is None:
            packer = cls.pack_varint
        return packer(cls.registry.encode_block(block))

    def unpack_block(self, unpacker=None):
        """
        Unpacks a block.
        """
        if unpacker is None:
            unpacker = self.unpack_varint
        return self.registry.decode_block(unpacker())

    # Slot --------------------------------------------------------------------

    @classmethod
    def pack_slot(cls, item=None, count=1, damage=0, tag=None):
        """
        Packs a slot.
        """

        if item is None:
            return cls.pack('h', -1)

        item_id = cls.registry.encode('minecraft:item', item)
        return cls.pack('hbh', item_id, count, damage) + cls.pack_nbt(tag)

    def unpack_slot(self):
        """
        Unpacks a slot.
        """

        slot = {}
        item_id = self.unpack('h')
        if item_id == -1:
            slot['item'] = None
        else:
            slot['item'] = self.registry.decode('minecraft:item', item_id)
            slot['count'] = self.unpack('b')
            slot['damage'] = self.unpack('h')
            slot['tag'] = self.unpack_nbt()

        return slot

    # NBT ---------------------------------------------------------------------

    @classmethod
    def pack_nbt(cls, tag=None):
        """
        Packs an NBT tag
        """

        if tag is None:
            # slower but more obvious:
            #   from quarry.types import nbt
            #   tag = nbt.TagRoot({})
            return b"\x00"

        return tag.to_bytes()

    def unpack_nbt(self):
        """
        Unpacks NBT tag(s).
        """

        from quarry.types import nbt
        return nbt.TagRoot.from_buff(self)


    # Entity metadata ---------------------------------------------------------

    @classmethod
    def pack_entity_metadata(cls, metadata):
        """
        Packs entity metadata.
        """
        out = b""
        for ty_key, val in metadata.items():
            ty, key = ty_key
            out += cls.pack('B', ty << 5 | key)
            if   ty == 0: out += cls.pack('b', val)
            elif ty == 1: out += cls.pack('h', val)
            elif ty == 2: out += cls.pack('i', val)
            elif ty == 3: out += cls.pack('f', val)
            elif ty == 4: out += cls.pack_string(val)
            elif ty == 5: out += cls.pack_slot(**val)
            elif ty == 6: out += cls.pack('iii', *val)
            elif ty == 7: out += cls.pack_rotation(*val)
            else: raise ValueError(f"Unknown entity metadata type: {ty:d}")
        out += cls.pack('B', 127)
        return out

    def unpack_entity_metadata(self):
        """
        Unpacks entity metadata.
        """
        metadata = {}
        while True:
            b = self.unpack('B')
            if b == 127:
                return metadata
            ty, key = b >> 5, b & 0x1F
            if ty == 0:
                val = self.unpack('b')
            elif ty == 1:
                val = self.unpack('h')
            elif ty == 2:
                val = self.unpack('i')
            elif ty == 3:
                val = self.unpack('f')
            elif ty == 4:
                val = self.unpack_string()
            elif ty == 5:
                val = self.unpack_slot()
            elif ty == 6:
                val = self.unpack('iii')
            elif ty == 7:
                val = self.unpack_rotation()
            else:
                raise ValueError(f"Unknown entity metadata type: {ty:d}")
            metadata[ty, key] = val

    # Direction ---------------------------------------------------------------

    @classmethod
    def pack_direction(cls, direction):
        """
        Packs a direction.
        """

        return cls.pack_varint(directions.index(direction))

    def unpack_direction(self):
        """
        Unpacks a direction.
        """

        return directions[self.unpack_varint()]

    # Rotation ----------------------------------------------------------------

    @classmethod
    def pack_rotation(cls, x, y, z):
        """
        Packs a rotation.
        """

        return cls.pack('fff', x, y, z)

    def unpack_rotation(self):
        """
        Unpacks a rotation
        """

        return self.unpack('fff')



import json

from twisted.internet import defer, reactor
from twisted.internet.defer import succeed
from twisted.python import failure
from twisted.web.client import Agent, error, Response, readBody
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from zope.interface import implementer


class HTTPException(Exception):
    def __init__(self, error_type, error_message):
        self.error_type = error_type
        self.error_message = error_message

    def __str__(self):
        return "%s: %s" % (self.error_type, self.error_message)


@implementer(IBodyProducer)
class BytesProducer:
    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


def request(url, timeout, err_type=Exception, expect_content=False, data=None):
    d0 = defer.Deferred()

    def _callback(response):
        def _callback2(body):
            if len(body):
                d0.callback(json.loads(body.decode('ascii')))
            else:
                d0.callback(None)
        d = readBody(response)
        d.addCallback(_callback2)
        return d

    def _errback(err):
        if isinstance(err.value, error.Error):
            if err.value.status == b"204":
                if expect_content:
                    err = failure.Failure(err_type(
                        "No Content",
                        "No content was returned by the server"))
                else:
                    d0.callback(None)
                    return
            else:
                data = json.loads(err.value.response.decode('ascii'))
                err = failure.Failure(err_type(
                    data['error'],
                    data['errorMessage']))
        d0.errback(err)

    agent = Agent(reactor)

    if data:
        d1 = agent.request(
            b'POST',
            url,
            Headers({"Content-Type": ["application/json"]}),
            BytesProducer(json.dumps(data).encode('ascii')),
        )
    else:
        d1 = agent.request(b'GET', url)

    d1.addCallbacks(_callback, _errback)

    return d0



import csv
import glob
import os.path
import re


def _load():
    default_protocol_version = 0
    minecraft_versions = {}
    packet_names = {}
    packet_idents = {}
    csv_paths = os.path.abspath(os.path.join(
        os.path.dirname(__file__),
        "packets",
        "*.csv"))
    for csv_path in glob.glob(csv_paths):
        match = re.match(r'(\d{4})_(.+)\.csv', os.path.basename(csv_path))
        if not match:
            continue

        protocol_version = int(match.group(1))
        minecraft_version = match.group(2)

        packet_ident = None
        last_section = None
        with open(csv_path) as csv_file:
            reader = csv.reader(csv_file)
            for i, record in enumerate(reader):
                # Skip header
                if i == 0:
                    continue

                # Extract fields
                protocol_mode = record[0]
                packet_direction = record[1]
                packet_name = record[2]

                section = (protocol_mode, packet_direction)
                if section != last_section:
                    packet_ident = 0
                last_section = section

                # Update default protocol version
                default_protocol_version = max(default_protocol_version,
                                               protocol_version)

                # Update minecraft versions
                minecraft_versions[protocol_version] = minecraft_version

                # Update the packet dictionaries
                key = [protocol_version, protocol_mode, packet_direction]
                packet_names[tuple(key + [packet_ident])] = packet_name
                packet_idents[tuple(key + [packet_name])] = packet_ident

                packet_ident += 1

    return (default_protocol_version, minecraft_versions,
            packet_names, packet_idents)

default_protocol_version, minecraft_versions, \
packet_names, packet_idents = _load()



from distutils.core import setup

setup(
    name='quarry',
    version='1.9.5',  # Also update doc/conf.py
    author='Barney Gale',
    author_email='barney@barneygale.co.uk',
    url='https://github.com/barneygale/quarry',
    license='MIT',
    description='Minecraft protocol library',
    long_description=open('README.rst').read(),
    install_requires=[
        'bitstring >= 3.1.0',
        'cached_property >= 1.2.0',
        'twisted >= 22.0.0',
        'cryptography >= 0.9',
        'pyOpenSSL >= 0.15.1',
        'service_identity >= 14.0.0',
    ],
    test_requires=[
        'pytest'
    ],
    packages=[
        "quarry",
        "quarry.data",
        "quarry.net",
        "quarry.types",
        "quarry.types.buffer",
    ],
    package_data={'quarry': [
        'data/packets/*.csv',
        'data/data_packs/*.nbt',
        'data/keys/*',
    ]},
)



from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from quarry.net.auth import PlayerPublicKey
from quarry.net.crypto import import_public_key
from quarry.types.buffer.v1_14 import Buffer1_14


class Buffer1_19(Buffer1_14):
    @classmethod
    def pack_player_public_key(cls, data: PlayerPublicKey):
        return cls.pack('Q', data.expiry) \
               + cls.pack_byte_array(data.key.public_bytes(Encoding.DER, PublicFormat.SubjectPublicKeyInfo)) \
               + cls.pack_byte_array(data.signature)

    def unpack_player_public_key(self):
        expiry = self.unpack('Q')
        key_bytes = self.unpack_byte_array()
        signature = self.unpack_byte_array()

        return PlayerPublicKey(expiry, import_public_key(key_bytes), signature)



from typing import List

from quarry.types.buffer.v1_19 import Buffer1_19
from quarry.types.chat import LastSeenMessage, SignedMessage, SignedMessageHeader, SignedMessageBody


class Buffer1_19_1(Buffer1_19):
    @classmethod
    def pack_last_seen_list(cls, entries: List[LastSeenMessage]):
        packed = cls.pack_varint(len(entries))

        if len(entries) > 5:
            from quarry.net.protocol import ProtocolError
            raise ProtocolError("Last seen list is too large")

        for entry in entries:
            packed = packed + cls.pack_last_seen_entry(entry)

        return packed

    def unpack_last_seen_list(self):
        seen_messages = []
        seen_messages_length = self.unpack_varint()

        if seen_messages_length > 5:
            from quarry.net.protocol import ProtocolError
            raise ProtocolError("Last seen list is too large")

        for i in range(seen_messages_length):
            seen_messages.append(self.unpack_last_seen_entry())

        return seen_messages

    @classmethod
    def pack_last_seen_entry(cls, entry: LastSeenMessage):
        return cls.pack_uuid(entry.sender) + cls.pack_byte_array(entry.signature)

    def unpack_last_seen_entry(self):
        return LastSeenMessage(self.unpack_uuid(), self.unpack_byte_array())

    @classmethod
    def pack_last_received(cls, entry: LastSeenMessage):
        return cls.pack_optional(cls.pack_last_seen_entry, entry)

    def unpack_last_received(self):
        return self.unpack_optional(self.unpack_last_seen_entry)

    @classmethod
    def pack_signed_message(cls, message: SignedMessage):
        return cls.pack_optional(cls.pack_byte_array, message.header.previous_signature) \
               + cls.pack_uuid(message.header.sender) \
               + cls.pack_byte_array(message.signature or b'') \
               + cls.pack_string(message.body.message) \
               + cls.pack_optional(cls.pack_chat, message.body.decorated_message) \
               + cls.pack('QQ', message.body.timestamp, message.body.salt) \
               + cls.pack_last_seen_list(message.body.last_seen) \
               + cls.pack_optional(cls.pack_chat, message.unsigned_content)

    def unpack_signed_message(self):
        previous_signature = self.unpack_optional(self.unpack_byte_array)
        uuid = self.unpack_uuid()
        signature = self.unpack_byte_array()
        message = self.unpack_string()
        decorated_message = self.unpack_optional(self.unpack_chat)
        timestamp = self.unpack('Q')
        salt = self.unpack('Q')
        last_seen = self.unpack_last_seen_list()
        unsigned_content = self.unpack_optional(self.unpack_chat)

        header = SignedMessageHeader(uuid, previous_signature)
        body = SignedMessageBody(message, timestamp, salt, decorated_message, last_seen)
        return SignedMessage(header, signature, 760, body, unsigned_content)



from twisted.internet import reactor, protocol, defer
from twisted.python import failure

from quarry.types.chat import Message
from quarry.net.protocol import Factory, Protocol, ProtocolError, \
    protocol_modes_inv
from quarry.net import auth, crypto


class ClientProtocol(Protocol):
    """This class represents a connection to a server"""

    recv_direction = "downstream"
    send_direction = "upstream"

    # Convenience functions ---------------------------------------------------

    def switch_protocol_mode(self, mode):
        self.check_protocol_mode_switch(mode)

        if mode in ("status", "login"):
            # Send handshake
            addr = self.transport.connector.getDestination()
            self.send_packet(
                "handshake",
                self.buff_type.pack_varint(self.protocol_version) +
                self.buff_type.pack_string(addr.host) +
                self.buff_type.pack('H', addr.port) +
                self.buff_type.pack_varint(
                    protocol_modes_inv[self.factory.protocol_mode_next]))

            # Switch buff type
            self.buff_type = self.factory.get_buff_type(self.protocol_version)

        self.protocol_mode = mode

        if mode == "status":
            # Send status request
            self.send_packet("status_request")

        elif mode == "login":
            # Send login start
            # TODO: Implement signatures/1.19.1 UUID sending
            if self.protocol_version >= 760:  # 1.19.1+
                self.send_packet("login_start",
                                 self.buff_type.pack_string(self.factory.profile.display_name),
                                 self.buff_type.pack("?", False),  # No signature as we haven't implemented them here
                                 self.buff_type.pack("?", False))  # No UUID as we haven't implemented them yet
            elif self.protocol_version == 759:  # 1.19
                self.send_packet("login_start",
                                 self.buff_type.pack_string(self.factory.profile.display_name),
                                 self.buff_type.pack("?", False))  # No signature as we haven't implemented them here
            else:
                # Send login start
                self.send_packet("login_start", self.buff_type.pack_string(
                    self.factory.profile.display_name))

    # Callbacks ---------------------------------------------------------------

    @defer.inlineCallbacks
    def connection_made(self):
        """Called when the connection is established"""
        super(ClientProtocol, self).connection_made()

        # Determine protocol version
        if self.factory.protocol_mode_next == "status":
            pass
        elif self.factory.force_protocol_version is not None:
            self.protocol_version = self.factory.force_protocol_version
        else:
            factory = PingClientFactory()
            factory.connect(self.remote_addr.host, self.remote_addr.port)
            self.protocol_version = yield factory.detected_protocol_version

        self.switch_protocol_mode(self.factory.protocol_mode_next)

    def auth_ok(self, data):
        """
        Called if the Mojang session server responds to our query. Note that
        this method does not indicate that the server accepted our session; in
        this case :meth:`player_joined` is called.
        """

        # Send encryption response
        p_shared_secret = crypto.encrypt_secret(
            self.public_key,
            self.shared_secret)
        p_verify_token = crypto.encrypt_secret(
            self.public_key,
            self.verify_token)

        # 1.7.x
        if self.protocol_version <= 5:
            pack_array = lambda d: self.buff_type.pack('h', len(d)) + d

        # 1.8.x
        else:
            pack_array = lambda d: self.buff_type.pack_varint(
                len(d), max_bits=16) + d

        # 1.19+
        if self.protocol_version >= 759:
            self.send_packet(
                "login_encryption_response",
                pack_array(p_shared_secret),
                self.buff_type.pack('?', True),  # Indicate we are still doing things the old way
                pack_array(p_verify_token))
        else:
            self.send_packet(
                "login_encryption_response",
                pack_array(p_shared_secret) +
                pack_array(p_verify_token))

        # Enable encryption
        self.cipher.enable(self.shared_secret)
        self.logger.debug("Encryption enabled")

    def player_joined(self):
        """
        Called when we join the game. If the server is in online mode, this
        means the server accepted our session.
        """
        Protocol.player_joined(self)
        self.logger.info("Joined the game.")

    def player_left(self):
        """Called when we leave the game."""
        Protocol.player_left(self)
        self.logger.info("Left the game.")

    def status_response(self, data):
        """
        If we're connecting in "status" mode, this is called when the server
        sends us information about itself.
        """
        self.close()

    # Packet handlers ---------------------------------------------------------

    def packet_status_response(self, buff):
        p_data = buff.unpack_json()
        self.status_response(p_data)

    def packet_login_plugin_request(self, buff):
        p_message_id = buff.unpack_varint()
        p_channel = buff.unpack_string()
        p_payload = buff.read()

        self.send_packet(
            "login_plugin_response",
            self.buff_type.pack_varint(p_message_id),
            self.buff_type.pack('?', False))

    def packet_login_disconnect(self, buff):
        p_data = buff.unpack_chat()
        self.logger.warn("Kicked: %s" % p_data)
        self.close()

    def packet_login_encryption_request(self, buff):
        p_server_id = buff.unpack_string()

        # 1.7.x
        if self.protocol_version <= 5:
            unpack_array = lambda b: b.read(b.unpack('h'))
        # 1.8.x
        else:
            unpack_array = lambda b: b.read(b.unpack_varint(max_bits=16))

        p_public_key = unpack_array(buff)
        p_verify_token = unpack_array(buff)

        if not self.factory.profile.online:
            raise ProtocolError("Can't log into online-mode server while using"
                                " offline profile")

        self.shared_secret = crypto.make_shared_secret()
        self.public_key = crypto.import_public_key(p_public_key)
        self.verify_token = p_verify_token

        # make digest
        digest = crypto.make_digest(
            p_server_id.encode('ascii'),
            self.shared_secret,
            p_public_key)

        # do auth
        deferred = self.factory.profile.join(digest)
        deferred.addCallbacks(self.auth_ok, self.auth_failed)

    def packet_login_success(self, buff):
        # 1.16.x
        if self.protocol_version >= 735:
            p_uuid = buff.unpack_uuid()
        # 1.15.x
        else:
            p_uuid = buff.unpack_string()
        p_display_name = buff.unpack_string()

        if self.protocol_version >= 759:
            buff.read()  # Properties

        self.switch_protocol_mode("play")
        self.player_joined()

    def packet_login_set_compression(self, buff):
        self.set_compression(buff.unpack_varint())

    def packet_set_compression(self, buff):
        self.set_compression(buff.unpack_varint())

    packet_disconnect = packet_login_disconnect


class SpawningClientProtocol(ClientProtocol):
    spawned = False

    def __init__(self, factory, remote_addr):
        # x, y, z, yaw, pitch
        self.pos_look = [0, 0, 0, 0, 0]

        super(SpawningClientProtocol, self).__init__(factory, remote_addr)

    # Send a 'player' packet every tick
    def update_player_inc(self):
        self.send_packet("player", self.buff_type.pack('?', True))

    # Sent a 'player position and look' packet every 20 ticks
    def update_player_full(self):
        self.send_packet(
            "player_position_and_look",
            self.buff_type.pack(
                'dddff?',
                self.pos_look[0],
                self.pos_look[1],
                self.pos_look[2],
                self.pos_look[3],
                self.pos_look[4],
                True))

    def packet_player_position_and_look(self, buff):
        p_pos_look = buff.unpack('dddff')

        # 1.7.x
        if self.protocol_version <= 5:
            p_on_ground = buff.unpack('?')
            self.pos_look = p_pos_look

        # 1.8.x
        else:
            p_flags = buff.unpack('B')

            for i in range(5):
                if p_flags & (1 << i):
                    self.pos_look[i] += p_pos_look[i]
                else:
                    self.pos_look[i] = p_pos_look[i]

            # 1.9.x
            if self.protocol_version > 47:
                teleport_id = buff.unpack_varint()

            if self.protocol_version > 754:
                dismount_vehicle = buff.unpack('?')

        # Send Player Position And Look

        # 1.7.x
        if self.protocol_version <= 5:
            self.send_packet("player_position_and_look", self.buff_type.pack(
                'ddddff?',
                self.pos_look[0],
                self.pos_look[1] - 1.62,
                self.pos_look[1],
                self.pos_look[2],
                self.pos_look[3],
                self.pos_look[4],
                True))

        # 1.8.x
        elif self.protocol_version <= 47:
            self.send_packet("player_position_and_look", self.buff_type.pack(
                'dddff?',
                self.pos_look[0],
                self.pos_look[1],
                self.pos_look[2],
                self.pos_look[3],
                self.pos_look[4],
                True))

        # 1.9.x
        else:
            self.send_packet("teleport_confirm", self.buff_type.pack_varint(
                teleport_id))

        if not self.spawned:
            self.spawn()

    def spawn(self):
        self.ticker.add_loop(1, self.update_player_inc)
        self.ticker.add_loop(20, self.update_player_full)
        self.spawned = True

    def packet_keep_alive(self, buff):
        self.send_packet('keep_alive', buff.read())


class ClientFactory(Factory, protocol.ClientFactory):
    protocol = ClientProtocol
    protocol_mode_next = "login"

    def __init__(self, profile=None):
        if profile is None:
            profile = auth.OfflineProfile()
        self.profile = profile

    def connect(self, host, port=25565):
        reactor.connectTCP(host, port, self, self.connection_timeout)


class PingClientProtocol(ClientProtocol):

    def status_response(self, data):
        self.close()
        detected_version = int(data["version"]["protocol"])
        if detected_version in self.factory.minecraft_versions:
            self.factory.detected_protocol_version.callback(detected_version)
        else:
            message = "Unsupported protocol version (%d)" % detected_version
            if 'description' in data:
                motd = Message(data['description'])
                message = "%s: %s" % (message, motd.to_string())
            self.factory.detected_protocol_version.errback(
                failure.Failure(ProtocolError(message)))


class PingClientFactory(ClientFactory):
    protocol = PingClientProtocol
    protocol_mode_next = "status"

    def __init__(self, profile=None):
        super(PingClientFactory, self).__init__(profile)
        self.detected_protocol_version = defer.Deferred()
